from .datadog_dashboard_summary import DatadogDashboardSummary
from .widget_metric import WidgetMetric
from .widget_options import WidgetOptions
from .query import Query
from .widget_aggregation_type import WidgetAggregationType
from .widget_aggregation_type_map import WidgetAggregationTypeMap
from .widget_options import WidgetOptions
from .widget_type import WidgetType
from .widget_type_map import WidgetTypeMap


import asyncio
from metrik.metrics.types import (
    Event,
    Interval,
    Rate
)
from typing import Dict, Union, Callable, Any, Literal
from .aws_lambda import (
    AWSLambda,
    AWSLambdaConfig
)
from .aws_timestream import (
    AWSTimestream,
    AWSTimestreamConfig
)
from .cloudwatch import (
    Cloudwatch,
    CloudwatchConfig
)
from .common import ConnectorTypes
from .csv import (
    CSV,
    CSVConfig
)
from .datadog import (
    Datadog,
    DatadogConfig
)
from .dogstatsd import (
    DogStatsD,
    DogStatsDConfig
)
from .json import (
    JSON,
    JSONConfig
)
from .kafka import (
    Kafka,
    KafkaConfig
)
from .mysql import (
    MySQL,
    MySQLConfig
)
from .postgres import (
    Postgres,
    PostgresConfig
)
from .prometheus import (
    Prometheus,
    PrometheusConfig
)
from .redis import (
    Redis,
    RedisConfig
)
from .s3 import (
    S3,
    S3Config
)
from .snowflake import (
    Snowflake,
    SnowflakeConfig
)
from .sqlite import (
    SQLite,
    SQLiteConfig
)
from .statsd import (
    StatsD,
    StatsDConfig
)
from .xml import (
    XML,
    XMLConfig
)


ConnectorConfig = Union[
    AWSLambdaConfig,
    AWSTimestreamConfig,
    CloudwatchConfig,
    CSVConfig,
    DatadogConfig,
    DogStatsDConfig,
    JSONConfig,
    KafkaConfig,
    MySQLConfig,
    PostgresConfig,
    PrometheusConfig,
    RedisConfig,
    S3Config,
    SnowflakeConfig,
    SQLiteConfig,
    StatsDConfig,
    XMLConfig
]

ConnectorInstance = Union[
    AWSLambda,
    AWSTimestream,
    Cloudwatch,
    CSV,
    Datadog,
    DogStatsD,
    JSON,
    Kafka,
    MySQL,
    Postgres,
    Prometheus,
    Redis,
    S3,
    Snowflake,
    SQLite,
    StatsD,
    XML
]


ConnectorTypeName = Union[
    Literal['aws-lambda'],
    Literal['aws-timestream'],
    Literal['cloudwatch'],
    Literal['csv'],
    Literal['datadog'],
    Literal['dogstatsd'],
    Literal['json'],
    Literal['kafka'],
    Literal['mysql'],
    Literal['postgres'],
    Literal['prometheus'],
    Literal['redis'],
    Literal['s3'],
    Literal['snowflake'],
    Literal['sqlite'],
    Literal['statsd'],
    Literal['xml']
]


class Connector:

    def __init__(self) -> None:
        
        self._connectors: Dict[
            ConnectorTypes,
            Callable[
                [ConnectorConfig],
                ConnectorInstance
            ]
        ] = {
            ConnectorTypes.AWSLambda: lambda config: AWSLambda(config),
            ConnectorTypes.AWSTimestream: lambda config: AWSTimestream(config),
            ConnectorTypes.Cloudwatch: lambda config: Cloudwatch(config),
            ConnectorTypes.CSV: lambda config: CSV(config),
            ConnectorTypes.Datadog: lambda config: Datadog(config),
            ConnectorTypes.DogStatsD: lambda config: DogStatsD(config),
            ConnectorTypes.JSON: lambda config: JSON(config),
            ConnectorTypes.Kafka: lambda config: Kafka(config),
            ConnectorTypes.MySQL: lambda config: MySQL(config),
            ConnectorTypes.Postgres: lambda config: Postgres(config),
            ConnectorTypes.Prometheus: lambda config: Prometheus(config),
            ConnectorTypes.Redis: lambda config: Redis(config),
            ConnectorTypes.S3: lambda config: S3(config),
            ConnectorTypes.Snowflake: lambda config: Snowflake(config),
            ConnectorTypes.SQLite: lambda config: SQLite(config),
            ConnectorTypes.StatsD: lambda config: StatsD(config),
            ConnectorTypes.XML: lambda config: XML(config)

        }



        self._configs: Dict[
            ConnectorTypes,
            Callable[
                [Dict[str, Any]],
                ConnectorConfig
            ]
        ] = {
            ConnectorTypes.AWSLambda: lambda config: AWSLambdaConfig(
                **config
            ),
            ConnectorTypes.AWSTimestream: lambda config: AWSTimestreamConfig(
                **config
            ),
            ConnectorTypes.Cloudwatch: lambda config: CloudwatchConfig(
                **config
            ),
            ConnectorTypes.CSV: lambda config: CSVConfig(
                **config
            ),
            ConnectorTypes.Datadog: lambda config: DatadogConfig(
                **config
            ),
            ConnectorTypes.DogStatsD: lambda config: DogStatsDConfig(
                **config
            ),
            ConnectorTypes.JSON: lambda config: JSONConfig(
                **config
            ),
            ConnectorTypes.Kafka: lambda config: KafkaConfig(
                **config
            ),
            ConnectorTypes.MySQL: lambda config: MySQLConfig(
                **config
            ),
            ConnectorTypes.Postgres: lambda config: PostgresConfig(
                **config
            ),
            ConnectorTypes.Prometheus: lambda config: PrometheusConfig(
                **config
            ),
            ConnectorTypes.Redis: lambda config: RedisConfig(
                **config
            ),
            ConnectorTypes.S3: lambda config: S3Config(
                **config
            ),
            ConnectorTypes.Snowflake: lambda config: SnowflakeConfig(
                **config
            ),
            ConnectorTypes.SQLite: lambda config: SQLiteConfig(
                **config
            ),
            ConnectorTypes.StatsD: lambda config: StatsDConfig(
                **config
            ),
            ConnectorTypes.XML: lambda config: XMLConfig(
                **config
            )
        }

        self._types: Dict[str, ConnectorTypes] = {
            'aws-lambda': ConnectorTypes.AWSLambda,
            'aws-timestream': ConnectorTypes.AWSTimestream,
            'cloudwatch': ConnectorTypes.Cloudwatch,
            'csv': ConnectorTypes.CSV,
            'datadog': ConnectorTypes.Datadog,
            'dogstatsd': ConnectorTypes.DogStatsD,
            'json': ConnectorTypes.JSON,
            'kafka': ConnectorTypes.Kafka,
            'mysql': ConnectorTypes.MySQL,
            'postgres': ConnectorTypes.Postgres,
            'prometheus': ConnectorTypes.Prometheus,
            'redis': ConnectorTypes.Redis,
            's3': ConnectorTypes.S3,
            'snowflake': ConnectorTypes.Snowflake,
            'sqlite': ConnectorTypes.SQLite,
            'statsd': ConnectorTypes.StatsD,
            'xml': ConnectorTypes.XML
        }

        self.selected_connector: Union[
            ConnectorInstance,
            None
        ] = None

        self.selected_connector_config: Union[
            ConnectorConfig,
            None
        ] = None

        self.selected_connector_type: Union[
            ConnectorTypes,
            None
        ] = None

        self._loop: Union[
            asyncio.AbstractEventLoop,
            None
        ] = None

    def select(
        self,
        connector_type: str,
        **kwargs
    ):
        
        self.selected_connector_type = self._types.get(connector_type)

        if self.selected_connector_type is None:
            raise Exception(
                f'Err. - no Connector of type - {connector_type} - exists.'
            )

        self.selected_connector_config = self._configs.get(
            self.selected_connector_type
        )(kwargs)

        self.selected_connector = self._connectors.get(
            self.selected_connector_type
        )(self.selected_connector_config)

    def load(
        self,
        source_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        return self._loop.run_until_complete(
            self._load(
                source_name,
                metric
            )
        )

    async def _load(
        self,
        source_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60
    ):
        await self.selected_connector.connect()
        result = await self.selected_connector.load(
            source_name,
            metric,
            timeout=timeout
        )

        await self.selected_connector.close()

        return result

    def send(
        self,
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60        
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        return self._loop.run_until_complete(
            self._send(
                destination_name,
                metric,
                timeout=timeout
            )
        )
    
    async def _send(
        self,
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60      
    ):
        await self.selected_connector.connect()
        await self.selected_connector.send(
            destination_name,
            metric,
            timeout=timeout
        )

        await self.selected_connector.close()



from typing import Any, Dict, Optional

from pydantic import BaseModel

from metrikflow.connectors.common.types import ConnectorTypes


class KafkaConfig(BaseModel):
    host: str='localhost:9092'
    client_id: str='metrikflow'
    group_id: str='metrikflow'
    metrics_partition: int=0
    compression_type: Optional[str]
    timeout: int=1000
    idempotent: bool=True
    options: Dict[str, Any]={}
    reporter_type: ConnectorTypes=ConnectorTypes.Kafka


import os
from pydantic import BaseModel
from metrik.connectors.common.types import ConnectorTypes


class SQLiteConfig(BaseModel):
    path: str=f'{os.getcwd()}/results.db'
    reporter_type: ConnectorTypes=ConnectorTypes.SQLite

    class Config:
        arbitrary_types_allowed = True


import uuid
from typing import Any, Dict, List, Union

from metrikflow.logging import MetrikflowLogger
from metrikflow.metrics import MetricStore
from metrikflow.metrics.types import Event, Interval, Rate

from .sqlite_config import SQLiteConfig

try:
    import sqlalchemy
    from sqlalchemy.ext.asyncio import create_async_engine
    from sqlalchemy.schema import CreateTable
    
    has_connector = True

except Exception:
    ASYNCIO_STRATEGY = None
    sqlalchemy = None
    CreateTable = None
    OperationalError = None
    has_connector = False



class SQLite:
    has_connector=has_connector

    def __init__(self, config: SQLiteConfig) -> None:
        self.path = f'sqlite+aiosqlite:///{config.path}'
        self.events_table_name = config.events_table
        self.metrics_table_name = config.metrics_table

        self.metadata = sqlalchemy.MetaData()

        self.database = None
        self._engine = None
        self._connection = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = MetrikflowLogger()
        self.logger.initialize()

        self._store = MetricStore()

    async def connect(self):
        self._engine = create_async_engine(self.path)

    async def load(
        self,
        source_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60
    ):
        table = sqlalchemy.Table(
            source_name,
            self.metadata,
            sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
            sqlalchemy.Column('metric_id', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('kind', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('group', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('timestamp', sqlalchemy.DateTime()),
            sqlalchemy.Column('value', sqlalchemy.Float())
        ) 


        metrics_data: List[Dict[str, Any]] = []

        async with self._engine.begin() as connection:
            
            await connection.execute(
                CreateTable(
                    table,
                    if_not_exists=True
                )
            )

            results = await connection.execute(
                table.select().where(
                    table.c.name == metric.name
                ).order_by(
                    table.c.timestamp.desc()
                ).limit(1)
            )

            for row in results.fetchall():
                metrics_data.append({
                    'metric_id': row.metric_id,
                    'name': row.name,
                    'kind': row.kind,
                    'group': row.group,
                    'timestamp': row.timestamp,
                    'value': row.value
                })

        return self._store.parse(
            metrics_data[-1]
        )

    async def send(
        self,
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60
    ):
        
        table = sqlalchemy.Table(
            destination_name,
            self.metadata,
            sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
            sqlalchemy.Column('metric_id', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('kind', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('group', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('timestamp', sqlalchemy.TEXT()),
            sqlalchemy.Column('value', sqlalchemy.Float())
        ) 
        
        async with self._engine.begin() as connection:

            await connection.execute(
                CreateTable(
                    table,
                    if_not_exists=True
                )
            )

            async with connection.begin() as transaction:

                metric_data = metric.dict()
                metric_data['kind'] = metric.kind.value

                await connection.execute(
                    table.insert().values(
                        **metric_data
                    )
                )   

                await transaction.commit()
    
    async def close(self):
        await self._engine.dispose()


import asyncio
import functools
import os
import uuid
from concurrent.futures import ThreadPoolExecutor
from typing import Dict, List, Union

from metrikflow.connectors.common.signals import add_signal_handler, handle_loop_stop
from metrikflow.logging import MetrikflowLogger
from metrikflow.metrics.types import Event, Interval, Rate
from metrikflow.metrics.types.base import MetricTypes

try:
    from prometheus_client import (
        CollectorRegistry,
        push_to_gateway,
    )
    from prometheus_client.core import REGISTRY
    from prometheus_client.exposition import basic_auth_handler

    from .prometheus_config import PrometheusConfig
    from .prometheus_metric import PrometheusMetric
    has_connector = True

except Exception:
    PrometheusConfig = None
    has_connector = False
    basic_auth_handler = lambda: None



class Prometheus:
    has_connector=has_connector

    def __init__(self, config: PrometheusConfig) -> None:
        self.pushgateway_address = config.pushgateway_address
        self.auth_request_method = config.auth_request_method
        self.auth_request_timeout = config.auth_request_timeout
        self.auth_request_data = config.auth_request_data
        self.username = config.username
        self.password = config.password
        self.namespace = config.namespace
        self.job_name = config.job_name

        self.registry = None
        self._auth_handler = None
        self._has_auth = False
        self._loop = asyncio.get_event_loop()


        self._metrics: Dict[str, PrometheusMetric] = {}

        self.metric_types_map = {
            MetricTypes.EVENT: 'gauge',
            MetricTypes.INTERVAL: 'gauge',
            MetricTypes.RATE: 'gauge'
        }

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = MetrikflowLogger()
        self.logger.initialize()

        self._executor = ThreadPoolExecutor(
            max_workers=os.cpu_count()
        )
        
    async def connect(self) -> None:

        self.registry = CollectorRegistry()
        REGISTRY.register(self.registry)

        if self.username and self.password:
            self._has_auth = True   

        add_signal_handler(
            self._loop,
            handle_loop_stop,
            self._executor,
            self._loop
        )   

    def _generate_auth(self) -> basic_auth_handler:
        return basic_auth_handler(
            self.pushgateway_address,
            self.auth_request_method,
            self.auth_request_timeout,
            {
                'Content-Type': 'application/json'
            },
            self.auth_request_data,
            username=self.username,
            password=self.password
        )
    
    async def _submit_to_pushgateway(
        self,
        pushgateway_address: str
    ):
        
        self.pushgateway_address = pushgateway_address

        if self._has_auth:

            await self._loop.run_in_executor(
                None,
                functools.partial(
                    push_to_gateway,
                    self.pushgateway_address,
                    job=self.job_name,
                    registry=self.registry,
                    handler=self._generate_auth
                )
            )

        else:
            await self._loop.run_in_executor(
                None,
                functools.partial(
                    push_to_gateway,
                    self.pushgateway_address,
                    job=self.job_name,
                    registry=self.registry
                )
            )

    async def load(
        self,
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60         
    ):
        raise Exception('Err. - Prometheus connector is send only.')
        

    async def send(
        self,
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60  
    ):
        tags_dict = metric.to_tags_dict()

        tags: List[str] = [
            f'{tag_name}:{tag_value}' for tag_name, tag_value in tags_dict.items()
        ]

        prometheus_metric = self._metrics.get(metric.name)

        if prometheus_metric is None:

            metric_type = self.metric_types_map.get(metric.kind, 'gauge')

            prometheus_metric = PrometheusMetric(
                metric.name,
                metric_type,
                metric_description=f'{metric.name} {metric.group}',
                metric_labels=tags,
                metric_namespace=self.namespace,
                registry=self.registry
            )
            prometheus_metric.create_metric()

        prometheus_metric.update(
            value=metric.value,
            labels=tags
        )

        await self._submit_to_pushgateway(
            destination_name
        )

    async def close(self):
        self._executor.shutdown(wait=False, cancel_futures=True)       


from pydantic import (
    BaseModel,
    StrictStr
)
from metrik.metrics.types import (
    Event,
    Interval,
    Rate
)
from .widget_options import WidgetOptions
from typing import (
    Optional, 
    Union,
    Literal
)


class WidgetMetric(BaseModel):
    widget_name: StrictStr
    widget_type: Literal[
        'change',
        'histogram',
        'query_value',
        'scatter_plot',
        'table',
        'timeseries'
    ]='timeseries'
    metric: Union[
        Event,
        Interval,
        Rate
    ]
    widget_size: Union[
        Literal['extra_small'],
        Literal['small'],
        Literal['medium'],
        Literal['large'],
        Literal['extra_large']
    ]='medium'
    widget_aggregator: Union[
        Literal['average'],
        Literal['total'],
        Literal['maximum'],
        Literal['minimum']
    ]='average'
    widget_aggregator_field: Optional[StrictStr]=None
    widget_options: Optional[WidgetOptions]=None



import json
import uuid
from typing import Union

from metrikflow.logging import MetrikflowLogger
from metrikflow.metrics import MetricStore
from metrikflow.metrics.types import Event, Interval, Rate

from .redis_config import RedisConfig

try:

    import aioredis
    has_connector = True

except Exception:
    aioredis = None
    has_connector = True


class Redis:
    has_connector=has_connector

    def __init__(self, config: RedisConfig) -> None:
        self.host = config.host
        self.base = 'rediss' if config.secure else 'redis'
        self.username = config.username
        self.password = config.password
        self.database = config.database

        self.connection = None
        self.pubsub = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = MetrikflowLogger()
        self.logger.initialize()

        self._store = MetricStore()

    async def connect(self):
        self.connection = await aioredis.from_url(
            f'{self.base}://{self.host}',
            username=self.username,
            password=self.password,
            db=self.database
        )

    async def load(
        self,
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60   
    ):
        data = await self.connection.get(f'{destination_name}_{metric.name}')
        return self._store.parse(
            json.loads(data)
        )

    async def send(
        self,
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60
    ):
        await self.connection.set(
            f'{destination_name}_{metric.name}',
            metric.json()
        )

    async def close(self):
        await self.connection.close()



from datadog_api_client.v1.model.change_widget_definition import ChangeWidgetDefinition
from datadog_api_client.v1.model.change_widget_definition_type import ChangeWidgetDefinitionType
from datadog_api_client.v1.model.change_widget_request import ChangeWidgetRequest
from datadog_api_client.v1.model.formula_and_function_metric_data_source import (
    FormulaAndFunctionMetricDataSource,
)
from datadog_api_client.v1.model.formula_and_function_metric_query_definition import (
    FormulaAndFunctionMetricQueryDefinition,
)
from datadog_api_client.v1.model.formula_and_function_metric_data_source import (
    FormulaAndFunctionMetricDataSource,
)
from datadog_api_client.v1.model.formula_and_function_response_format import (
    FormulaAndFunctionResponseFormat
)
from datadog_api_client.v1.model.widget_sort import WidgetSort
from datadog_api_client.v1.model.widget_formula import WidgetFormula
from datadog_api_client.v1.model.widget_order_by import WidgetOrderBy
from datadog_api_client.v1.model.widget_formula_limit import WidgetFormulaLimit
from datadog_api_client.v1.model.widget_change_type import WidgetChangeType
from datadog_api_client.v1.model.widget_compare_to import WidgetCompareTo
from datadog_api_client.v1.model.widget_text_align import WidgetTextAlign
from datadog_api_client.v1.model.widget_time import WidgetTime
from metrik.dashboards.datadog.models import (
    WidgetOptions,
    WidgetType
)
from metrik.dashboards.datadog.widgets.widget_types.base.base_widget_factory import BaseWidgetFactory
from typing import List


class ChangeFactory(BaseWidgetFactory[ChangeWidgetDefinition]):
    
    def __init__(self) -> None:
        super().__init__(WidgetType.CHANGE)

    def create(
        self,
        widget_name: str,
        options: WidgetOptions
    ):
        
        queries = self.queries[widget_name]
        
        comparator = WidgetCompareTo(
            value=options.comparison_window
        )

        change_type = WidgetChangeType(
            value=options.comparison_type
        )

        sort_direction = WidgetSort(
            value=options.order
        )

        formulas: List[WidgetFormula]= []
        for query in queries:
            formulas.extend([
                WidgetFormula(
                    formula=f'{options.comparison_window}({query.to_query_name()})'
                ),
                WidgetFormula(
                    formula=query.to_query_name(),
                    limit=WidgetFormulaLimit(
                        count=options.limit,
                        order=options.to_datadog_sort_order()
                    )
                )
            ])

        change = ChangeWidgetDefinition(
            title=widget_name,
            title_size='16',
            title_align=WidgetTextAlign.LEFT,
            type=ChangeWidgetDefinitionType.CHANGE,
            time=WidgetTime(
                live_span=options.to_datadog_live_span()
            ),
            requests=[
                ChangeWidgetRequest(
                    formulas=formulas,
                    queries=[
                        FormulaAndFunctionMetricQueryDefinition(
                            data_source=FormulaAndFunctionMetricDataSource.METRICS,
                            name=query.to_query_name(),
                            aggregator=query.formula_type,
                            query=query.to_query()
                        ) if query.aggregation_field else FormulaAndFunctionMetricQueryDefinition(
                            data_source=FormulaAndFunctionMetricDataSource.METRICS,
                            name=query.to_query_name(),
                            query=query.to_query()
                        )   for query in queries
                    ],
                    compare_to=comparator,
                    change_type=change_type,
                    order_dir=sort_direction,
                    order_by=options.to_widget_order_by(),
                    response_format=FormulaAndFunctionResponseFormat.SCALAR,
                    show_present=options.display_current_values
                )
            ],
        )

        self._widgets[widget_name] = change

        return change




from collections import OrderedDict
from typing import List, Union

from tabulate import tabulate

from metrikflow.logging import MetrikflowLogger
from metrikflow.metrics.types import Event, Interval, Rate
from metrikflow.tools.formatting import seconds_to_human_string


class MetrikTable:

    def __init__(
        self,
        metrics: List[
            Union[
                Event, 
                Interval, 
                Rate
            ]
        ]
    ) -> None:
        
        self.table: Union[str, None] = None

        self.metrics = metrics
        self.table_rows: List[OrderedDict] = []

        self.logger = MetrikflowLogger()
        self.logger.initialize()

    def generate_table(self):

        for metric in self.metrics:
            table_row = OrderedDict()

            metric_data = metric.to_data_dict()

            for field_name in metric.to_column_names():
                header_name = field_name.replace('_', ' ')
                metric_value = metric_data[field_name]

                if field_name == 'value' and isinstance(metric, Interval):
                    metric_value = seconds_to_human_string(metric_value)

                table_row[header_name] = metric_value

            for tag_set in metric.tags:
                for tag_name, tag_value in tag_set.items():
                    table_row[tag_name] = tag_value

            self.table_rows.append(table_row)

        self.table = tabulate(
            self.table_rows,
            headers='keys',
            missingval='None',
            tablefmt="simple",
            floatfmt=".2f"
        )

        self.logger.console.sync.info('\nMetrics:\n')
        self.logger.console.sync.info(f'''{self.table}\n''')



import asyncio
import signal
from typing import Callable, Awaitable


def add_signal_handler(
        loop: asyncio.AbstractEventLoop,
        handler: Callable[
            ...,
            Awaitable[None]
        ],
        *args
):
    for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
        loop.add_signal_handler(
            getattr(signal, signame),
            lambda signame=signame: handler(
                signame,
                *args
            )
        )



import time
from typing import Union, Any
from datetime import datetime
from metrik.metrics.types import (
    Event,
    Interval,
    Rate
)


class AWSTimestreamRecord:

    def __init__(
        self, 
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        session_id: str
    ) -> None:

        measure_value_type = None
        value = metric.serialize_value()

        if isinstance(value, float):
            measure_value_type = "DOUBLE"

        elif isinstance(value, str):
            measure_value_type = "VARCHAR"
        
        elif isinstance(value, bool):
            measure_value_type = "BOOLEAN"

        elif isinstance(value, int):
            measure_value_type = "BIGINT"

        elif isinstance(value, datetime):
            value = int(value.timestamp())
            measure_value_type = "TIMESTAMP"

        self.record_kind = metric.kind.value
        self.record_name = metric.name
        self.field = 'value'

        self.time = str(int(round(time.time() * 1000)))
        self.time_unit = 'MILLISECONDS'
        self.dimensions = [
            {
                "Name": "metric_id",
                "Value": metric.metric_id
            },
            {
                "Name": "name",
                "Value": metric.name
            },
            {
                "Name": "kind",
                "Value": metric.kind.value
            },
            {
                "Name": "timestamp",
                "Value": metric.to_date_string()
            },
            {
                "Name": "session_id",
                "Value": session_id
            }
        ]
        self.measure_name = f'{metric.name}_{metric.group}_{session_id}'
        self.measure_value = str(value)
        self.measure_value_type = measure_value_type

    def to_dict(self):
        return  {
            "Time": self.time,
            "TimeUnit": self.time_unit,
            "Dimensions": self.dimensions,
            "MeasureName": self.measure_name,
            "MeasureValue": self.measure_value,
            "MeasureValueType": self.measure_value_type
        }
    
    @classmethod
    def parse_type(
        cls,
        column_type: str,
        value: str
    ):
        match column_type:
            case 'VARCHAR':
                return value
            
            case 'BOOLEAN':
                return True if value.lower() == 'true' else False

            case 'BIGINT' | 'INTEGER' | 'TIME':
                return int(value)
            
            case 'DOUBLE':
                return float(value)
            
            case 'DATE':
                return datetime.strptime(
                    value,
                    '%Y-%m-%d'
                )
            
            case 'TIMESTAMP':
                return datetime.strptime(
                    value,
                    '%Y-%m-%dT%H:%M:%S.%f.%z'
                )
            
            case _:
                return value



from datadog_api_client.v1.model.timeseries_widget_definition import TimeseriesWidgetDefinition
from datadog_api_client.v1.model.timeseries_widget_definition_type import TimeseriesWidgetDefinitionType
from datadog_api_client.v1.model.timeseries_widget_request import (
    TimeseriesWidgetRequest,
)
from datadog_api_client.v1.model.timeseries_widget_legend_column import TimeseriesWidgetLegendColumn
from datadog_api_client.v1.model.timeseries_widget_legend_layout import TimeseriesWidgetLegendLayout
from datadog_api_client.v1.model.widget_text_align import WidgetTextAlign
from datadog_api_client.v1.model.formula_and_function_metric_query_definition import (
    FormulaAndFunctionMetricQueryDefinition,
)
from datadog_api_client.v1.model.formula_and_function_metric_data_source import (
    FormulaAndFunctionMetricDataSource,
)
from datadog_api_client.v1.model.formula_and_function_response_format import (
    FormulaAndFunctionResponseFormat
)
from datadog_api_client.v1.model.widget_formula_style import WidgetFormulaStyle
from datadog_api_client.v1.model.widget_time import WidgetTime
from datadog_api_client.v1.model.widget_formula import WidgetFormula
from metrik.dashboards.datadog.models import (
    WidgetOptions,
    WidgetType
)
from metrik.dashboards.datadog.widgets.widget_types.base.base_widget_factory import BaseWidgetFactory


class TimeseriesFactory(BaseWidgetFactory[TimeseriesWidgetDefinition]):

    def __init__(self) -> None:
        super().__init__(WidgetType.TIMESERIES)
    
    def create(
        self,
        widget_name: str,
        options: WidgetOptions
    ):
        
        queries = self.queries[widget_name]


        timeseries = TimeseriesWidgetDefinition(
            title=widget_name,
            title_size="16",
            title_align=WidgetTextAlign.LEFT,
            show_legend=True,
            type=TimeseriesWidgetDefinitionType.TIMESERIES,
            legend_columns=[
                TimeseriesWidgetLegendColumn.AVG,
                TimeseriesWidgetLegendColumn.SUM,
                TimeseriesWidgetLegendColumn.MAX,
                TimeseriesWidgetLegendColumn.MIN,
                TimeseriesWidgetLegendColumn.VALUE
            ],
            legend_layout=TimeseriesWidgetLegendLayout.AUTO,
            legend_size='12',
            time=WidgetTime(
                live_span=options.to_datadog_live_span()
            ),
            requests=[
                TimeseriesWidgetRequest(
                    formulas=[
                        WidgetFormula(
                            formula=query.to_query_name()
                        ) for query in queries
                    ],
                    queries=[
                        FormulaAndFunctionMetricQueryDefinition(
                            data_source=FormulaAndFunctionMetricDataSource.METRICS,
                            name=query.to_query_name(),
                            query=query.to_query()
                        )  for query in queries
                    ],
                    response_format=FormulaAndFunctionResponseFormat.TIMESERIES
                )
            ]
        )

        self._widgets[widget_name] = timeseries

        return timeseries


import os
import json
import datetime
from typing import (
    Literal, 
    Union, 
    Dict, 
    Any,
    List
)
from metrik.connectors import Connector
from metrik.logging.table import MetrikTable
from metrik.metrics import (
    MetricStore
)


def record_metric(
    name: str,
    kind: Union[
        Literal['event'],
        Literal['interval'],
        Literal['rate']
    ],
    group: str=None,
    value: float=None,
    load: str=None,
    send: str=None,
    config_path: str=None,
    timeout: Union[int, float]=60,
    tags: str=None
):
    
    if send is None:
        metric_send_location = os.path.join(
            os.getcwd(),
            f'{name}_{kind}.json'
        )

        metric_format = 'json'

    else:
        metric_format, metric_send_location = send.split(':', maxsplit=1)

    metrik_config = {}
    if os.path.exists(config_path):
        with open(config_path) as metrik_config_file:
            metrik_config = json.load(metrik_config_file)

    metrik_config['config_path'] = config_path

    store = MetricStore()
    connector = Connector()


    if load:
        load_connector_config: Dict[str, Any] = metrik_config.get('loader', {})

        metric_format, metric_load_location = load.split(':', maxsplit=1)

        metric_load_query = store.create_load_query(
            name=name,
            kind=kind,
            group=group,
            timestamp=datetime.datetime.now(datetime.timezone.utc),
            value=value
        )

        connector.select(
            metric_format,
            **load_connector_config
        )

        loaded_metric = connector.load(
            metric_load_location,
            metric_load_query,
            timeout=timeout
        )

        if loaded_metric:
            store.record(
                name=loaded_metric.name,
                kind=kind,
                group=loaded_metric.group,
                timestamp=loaded_metric.timestamp,
                value=loaded_metric.value
            )
    
    metric_tags: Union[
        List[
            Dict[str, str]
        ],
        None
    ] = None
    if tags:
        metric_tags = []
        tag_strings = tags.split(',')
        
        for tag_string in tag_strings:
            tag_name, tag_value = tag_string.split(':')
            metric_tags.append({
                tag_name: tag_value
            })

    metric = store.record(
        name=name,
        kind=kind,
        group=group,
        timestamp=datetime.datetime.now(datetime.timezone.utc),
        value=value,
        tags=metric_tags
    )


    send_connector_config: Dict[str, Any] = metrik_config.get(
        'sender', 
        {}
    )

    connector.select(
        metric_format,
        **send_connector_config
    )

    connector.send(
        metric_send_location,
        metric,
        timeout=timeout
    )

    table = MetrikTable([metric])

    table.generate_table()



from datadog_api_client.v1.model.scatter_plot_widget_definition import ScatterPlotWidgetDefinition
from datadog_api_client.v1.model.scatter_plot_widget_definition_type import ScatterPlotWidgetDefinitionType
from datadog_api_client.v1.model.scatter_plot_widget_definition_requests import ScatterPlotWidgetDefinitionRequests
from datadog_api_client.v1.model.scatterplot_table_request import ScatterplotTableRequest
from datadog_api_client.v1.model.scatterplot_widget_formula import ScatterplotWidgetFormula
from datadog_api_client.v1.model.formula_and_function_metric_data_source import (
    FormulaAndFunctionMetricDataSource,
)
from datadog_api_client.v1.model.formula_and_function_metric_query_definition import (
    FormulaAndFunctionMetricQueryDefinition,
)
from datadog_api_client.v1.model.formula_and_function_metric_data_source import (
    FormulaAndFunctionMetricDataSource,
)
from datadog_api_client.v1.model.formula_and_function_response_format import (
    FormulaAndFunctionResponseFormat
)
from datadog_api_client.v1.model.widget_axis import WidgetAxis
from datadog_api_client.v1.model.widget_style import WidgetStyle
from datadog_api_client.v1.model.widget_time import WidgetTime
from datadog_api_client.v1.model.widget_text_align import WidgetTextAlign
from metrik.dashboards.datadog.models import (
    WidgetOptions,
    WidgetType
)
from metrik.dashboards.datadog.widgets.widget_types.base.base_widget_factory import BaseWidgetFactory


class ScatterPlotFactory(BaseWidgetFactory[ScatterPlotWidgetDefinition]):

    def __init__(self) -> None:
        super().__init__(WidgetType.SCATTER_PLOT)

    def create(
        self,
        widget_name: str,
        options: WidgetOptions
    ):   
        queries = self.queries[widget_name]
  
        scatter_plot = ScatterPlotWidgetDefinition(
            title=widget_name,
            title_size='16',
            title_align=WidgetTextAlign.LEFT,
            type=ScatterPlotWidgetDefinitionType.SCATTERPLOT,
            time=WidgetTime(
                live_span=options.to_datadog_live_span()
            ),
            xaxis=WidgetAxis(
                label=options.x_axis_name,
                max="auto",
                include_zero=True,
                scale="linear",
                min="auto",
            ),
            yaxis=WidgetAxis(
                label=options.y_axis_name,
                max="auto",
                include_zero=True,
                scale="linear",
                min="auto",
            ),
            requests=[
                ScatterPlotWidgetDefinitionRequests(
                    table=ScatterplotTableRequest(
                        formulas=[
                            ScatterplotWidgetFormula(
                                formula=query.to_query_name(),
                                dimension='x' if query.metric in options.x_axis_metrics else 'y',
                            ) for  query in queries
                        ],
                        queries=[
                            FormulaAndFunctionMetricQueryDefinition(
                            data_source=FormulaAndFunctionMetricDataSource.METRICS,
                            name=query.to_query_name(),
                            aggregator=query.formula_type,
                            query=query.to_query()
                        ) if query.aggregation_field else FormulaAndFunctionMetricQueryDefinition(
                            data_source=FormulaAndFunctionMetricDataSource.METRICS,
                            name=query.to_query_name(),
                            query=query.to_query()
                        )  for query in queries
                        ],
                        response_format=FormulaAndFunctionResponseFormat.TIMESERIES
                    )
                )
            ]
        )

        self._widgets[widget_name] = scatter_plot

        return scatter_plot


import asyncio
import collections
import collections.abc
import datetime
import functools
import os
import uuid
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Dict, List, TextIO, Union

from metrikflow.connectors.common.signals import (
    add_signal_handler,
    handle_file_stop,
    handle_loop_stop,
)
from metrikflow.logging import MetrikflowLogger
from metrikflow.metrics import MetricStore
from metrikflow.metrics.types import Event, Interval, Rate

from .xml_config import XMLConfig

try:
    from dicttoxml import dicttoxml
    from xmltodict import parse
    has_connector=True

except Exception:
    dicttoxml = object
    has_connector=False

collections.Iterable = collections.abc.Iterable


MetricRecord = Dict[str, Union[int, float, str]]
MetricRecordGroup = Dict[str, List[MetricRecord]]
MetricRecordCollection = Dict[str, MetricRecord]
    

class XML:
    has_connector=has_connector
    
    def __init__(self, config: XMLConfig) -> None:

        self._executor = ThreadPoolExecutor(
            os.cpu_count()
        )
        self._loop: asyncio.AbstractEventLoop = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = MetrikflowLogger()
        self.logger.initialize()

        self.metrics_file: TextIO = None
        self._store = MetricStore()

        self.write_mode = 'w' if config.overwrite else 'a'

    async def connect(self):
        self._loop = asyncio._get_running_loop()
      
        add_signal_handler(
            self._loop,
            handle_loop_stop,
            self._executor,
            self._loop
        )

    async def load(
        self,
        source_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60 
    ):
        absolute_path = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                os.path.abspath,
                source_name
            )
        )

        file_exists = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                os.path.exists,
                absolute_path
            )
        )

        if file_exists is False:
            return None

        self.metrics_file = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                open,
                absolute_path,
                'r'
            )
        )

        add_signal_handler(
            self._loop,
            handle_file_stop,
            self.metrics_file
        )

        data = await self._loop.run_in_executor(
            self._executor,
            self.metrics_file.read
        )

        metrics_data = parse(data)

        metrics: List[
            Dict[str, Any]
        ] = list(
            sorted(
                metrics_data.get(metric.name, []),
                key=lambda metric: datetime.datetime.strptime(
                    metric.get('timestamp'),
                    '%Y-%m-%dT%H:%M:%S.%f.%z'
                )
            )
        )

        if len(metrics) < 1:
            return None

        return self._store.parse(
            metrics[-1]
        )

    async def send(
        self,
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60  
    ):
        absolute_path = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                os.path.abspath,
                destination_name
            )
        )

        file_exists = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                os.path.exists,
                absolute_path
            )
        )


        data = metric.to_data_dict()

        self.metrics_file = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                open,
                absolute_path,
                'r+'
            )
        )

        add_signal_handler(
            self._loop,
            handle_file_stop,
            self.metrics_file
        )

        if file_exists:
            
            data: str = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.metrics_file.read
                )
            )

            metrics_data: Dict[
                str, 
                List[Dict[str, Any]]
            ] = parse(data)

            metrics = metrics_data.get(metric.name, [])
            metrics.append(data)
            metrics_data[metric.name] = metrics

            xml_data = dicttoxml(metrics_data)

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.metrics_file.write,
                    xml_data
                )
            )

        else:
            xml_data = dicttoxml({
                metric.name: [
                    metric.to_data_dict()
                ]
            })

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.metrics_file.write,
                    xml_data
                )
            )

    async def close(self):

        await self._loop.run_in_executor(
            self._executor,
            self.metrics_file.close
        )

        self._executor.shutdown(wait=False, cancel_futures=True)


from typing import Optional
from pydantic import BaseModel
from metrik.connectors.common.types import ConnectorTypes

class SnowflakeConfig(BaseModel):
    username: str
    password: str
    organization_id: str
    account_id: str
    private_key: Optional[str]
    warehouse: str
    database: str
    database_schema: str='PUBLIC'
    connect_timeout: int=30
    reporter_type: ConnectorTypes=ConnectorTypes.Snowflake

    class Config:
        arbitrary_types_allowed = True


import asyncio
import csv
import datetime
import functools
import os
import uuid
from concurrent.futures import ThreadPoolExecutor
from typing import Dict, List, TextIO, Union

from metrikflow.connectors.common.signals import (
    add_signal_handler,
    handle_file_stop,
    handle_loop_stop,
)
from metrikflow.logging import MetrikflowLogger
from metrikflow.metrics import MetricStore
from metrikflow.metrics.types import Event, Interval, Rate

from .csv_config import CSVConfig

has_connector = True


class CSV:
    has_connector=has_connector

    def __init__(self, config: CSVConfig) -> None:
        self._executor = ThreadPoolExecutor(
            max_workers=os.cpu_count()
        )

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = MetrikflowLogger()
        self.logger.initialize()

        self._csv_reader: Union[csv.DictReader, None] = None
        self._csv_writer: Union[csv.DictWriter, None] = None

        self._loop: asyncio.AbstractEventLoop = None

        self.metrics_file: TextIO = None 

        self.write_mode = 'w' if config.overwrite else 'a'
        self._store = MetricStore()

    async def connect(self):
        self._loop = asyncio._get_running_loop()

        add_signal_handler(
            self._loop,
            handle_loop_stop,
            self._executor,
            self._loop
        )

    async def load(
        self,
        source_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60       
    ):
        absolute_path = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                os.path.abspath,
                source_name
            )
        )

        file_exists = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                os.path.exists,
                absolute_path
            )
        )

        if file_exists is False:
            return None

        self.metrics_file = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                open,
                absolute_path,
                'r'
            )
        )

        add_signal_handler(
            self._loop,
            handle_file_stop,
            self.metrics_file
        )

        lines = await self._loop.run_in_executor(
            self._executor,
            self.metrics_file.readlines
        )

        self._csv_reader = csv.DictReader(
            lines, 
            metric.to_column_names()
        )

        metric_records: List[
            Dict[str, any]
        ] = list(
            sorted(
                [
                    line for line in self._csv_reader if line.get(
                        'name'
                    ) == metric.name
                ],
                key=lambda record: datetime.datetime.strptime(
                    record.get('timestamp'),
                    '%Y-%m-%dT%H:%M:%S.%f.%z'
                )
            )
        )

        if len(metric_records) < 1:
            return None

        return self._store.parse(
            metric_records[-1]
        )

    async def send(
        self,
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60
    ):
        
        absolute_path = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                os.path.abspath,
                destination_name
            )
        )


        file_exists = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                os.path.exists,
                absolute_path
            )
        )
        
        self.metrics_file = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                open,
                absolute_path,
                'a+'
            )
        )

        add_signal_handler(
            self._loop,
            handle_file_stop,
            self.metrics_file
        )

        self._csv_writer = csv.DictWriter(
            self.metrics_file,
            metric.to_column_names()
        )


        if file_exists is False:
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self._csv_writer.writeheader
                )
            )

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._csv_writer.writerow,
                metric.dict()
            )
        )

    async def close(self):
        await self._loop.run_in_executor(
            self._executor,
            self.metrics_file.close
        )

        self._executor.shutdown(wait=False, cancel_futures=True)


from typing import Dict
from pydantic import BaseModel
from metrik.connectors.common.types import ConnectorTypes


class AWSTimestreamConfig(BaseModel):
    aws_access_key_id: str
    aws_secret_access_key: str
    region_name: str
    database_name: str
    retention_options: Dict[str, int] = {
        "MemoryStoreRetentionPeriodInHours": 1,
        "MagneticStoreRetentionPeriodInDays": 365,
    }
    reporter_type: ConnectorTypes=ConnectorTypes.AWSTimestream


import json
import time
import uuid
from typing import Any, Dict, List, Union

from metrikflow.logging import MetrikflowLogger
from metrikflow.metrics import MetricStore
from metrikflow.metrics.types import Event, Interval, Rate

from .kafka_config import KafkaConfig

try:

    from aiokafka import AIOKafkaConsumer, AIOKafkaProducer, ConsumerRecord
    has_connector = True

except Exception:
    AIOKafkaProducer = None
    AIOKafkaConsumer = None
    ConsumerRecord = None
    has_connector = False


class Kafka:
    has_connector=has_connector

    def __init__(self, config: KafkaConfig) -> None:
        self.host = config.host
        self.client_id = config.client_id

        self.metrics_partition = config.metrics_partition
        self.group_id = config.group_id

        self.compression_type = config.compression_type
        self.timeout = config.timeout
        self.enable_idempotence = config.idempotent or True
        self.options: Dict[str, Any] = config.options or {}

        self._client: Union[
            AIOKafkaConsumer,
            None
        ] = None
        self._producer: Union[
            AIOKafkaProducer,
            None
        ] = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = MetrikflowLogger()
        self.logger.initialize()

        self._store = MetricStore()

    async def connect(self):

        self._client = AIOKafkaConsumer(
            bootstrap_servers=self.host,
            client_id=self.client_id,
            group_id=self.group_id,
            request_timeout_ms=self.timeout,
            **self.options
        )

        self._producer = AIOKafkaProducer(
            bootstrap_servers=self.host,
            client_id=self.client_id,
            compression_type=self.compression_type,
            request_timeout_ms=self.timeout,
            enable_idempotence=self.enable_idempotence,
            **self.options
        )

        await self._client.start()
        await self._producer.start()

    async def load(
        self,
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60
    ):
        await self._client.subscribe(
            topics=(
                destination_name
            )
        )

        messages: List[ConsumerRecord] = []
        elapsed = 0
        start = time.time()

        async for message in self._client:
            if message.key == metric.name:

                messages.append(message)

            elapsed = time.time() - start

            if elapsed > timeout:
                break

        messages = list(sorted(
            messages,
            key=lambda message: message.timestamp
        ))

        return self._store.parse(
            json.loads(
                messages[-1].value.decode(
                    'utf-8'
                )
            )
        )

    async def send(
        self,
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60
    ):
        
        if self._producer:
            
            batch = self._producer.create_batch()
            batch.append(
                value=metric.json().encode('utf-8'),
                timestamp=metric.timestamp.timestamp(), 
                key=bytes(metric.name, 'utf-8')
            )

            await self._producer.send_batch(
                batch,
                destination_name,
                partition=self.metrics_partition
            )

    async def close(self):
        await self._producer.stop()



import asyncio
import functools
import os
import uuid
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Dict, List, Union

from metrikflow.connectors.common.signals import add_signal_handler, handle_loop_stop
from metrikflow.logging import MetrikflowLogger
from metrikflow.metrics import MetricStore
from metrikflow.metrics.types import Event, Interval, Rate

from .aws_timestream_config import AWSTimestreamConfig
from .aws_timestream_record import AWSTimestreamRecord

try:

    import boto3
    has_connector = True
except Exception:
    has_connector = False

class AWSTimestream:
    has_connector=has_connector

    def __init__(self, config: AWSTimestreamConfig) -> None:
        self.aws_access_key_id = config.aws_access_key_id
        self.aws_secret_access_key = config.aws_secret_access_key
        self.region_name = config.region_name

        self.database_name = config.database_name

        self.retention_options = config.retention_options
        self.session_uuid = str(uuid.uuid4())

        self._executor = ThreadPoolExecutor(
            max_workers=os.cpu_count()
        )
        self.query_client = None
        self.write_client = None
        self._loop = asyncio.get_event_loop()
        self.metadata_string: str = None

        self.logger = MetrikflowLogger()
        self.logger.initialize()

        self._store = MetricStore()

    async def connect(self):

        add_signal_handler(
            self._loop,
            handle_loop_stop,
            self._executor,
            self._loop
        )

        self.query_client = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                boto3.client,
                'timestream-query',
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                region_name=self.region_name
            )
        )


        self.write_client = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                boto3.client,
                'timestream-write',
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                region_name=self.region_name
            )
        )

        try:

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.write_client.create_database,
                    DatabaseName=self.database_name
                )
            )
 
        except Exception:
            pass

    async def load(
        self,
        source_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60
    ):
        
        try:
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.write_client.create_table,
                    DatabaseName=self.database_name,
                    TableName=source_name,
                    RetentionProperties=self.retention_options
                )
            )

        except Exception:
           pass

        query = f'SELECT * FROM {self.database_name}.{source_name} WHERE name={metric.name} ORDER BY time DESC'

        try:

            response: Dict[
                str, 
                List[Dict[str, Any]]
            ] = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.query_client.query,
                    QueryString=query
                )
            )

            column_info = response.get('ColumnInfo', [])
            
            rows = response.get('Rows', [])

            if len(rows) == 0:
                return None
            
            metric_rows: List[Dict[str, Any]] = {}

            for row in rows:
            
                metric_rows: List[
                    Dict[str, Any]
                ] = row.get('Data', [])
                
                metric_data: Dict[str, Any] = {}

                for column, data in zip(column_info, metric_rows):

                    name = column.get('Name')
                    column_type_data: Dict[str, str] = column.get('Type')
                    column_type = column_type_data.get('ScalarType')

                    value = AWSTimestreamRecord.parse_type(
                        column_type,
                        data.get("ScalarValue")
                    )

                    metric_data[name] = value

                metric_rows.append(metric_data)


            metrics: List[Dict[str, Any]] = list(
                sorted(
                    metric_rows,
                    key=lambda metric_data: metric_data.get(
                        'timestamp'
                    )
                )
            )
                
            metric = self._store.parse(
                metrics[-1]
            )

        except Exception:
            pass

        return metric
        
    async def send(
        self, 
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60
    ):

        try:
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.write_client.create_table,
                    DatabaseName=self.database_name,
                    TableName=destination_name,
                    RetentionProperties=self.retention_options
                )
            )

        except Exception:
           pass

        timestream_record = AWSTimestreamRecord(
            metric,
            self.session_uuid
        )

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.write_client.write_records,
                DatabaseName=self.database_name,
                TableName=destination_name,
                Records=[
                    timestream_record.to_dict()
                ],
                CommonAttributes={}
            )
        )

    async def close(self):
        self._executor.shutdown(wait=False, cancel_futures=True)


import os
from pydantic import BaseModel
from metrik.connectors.common.types import ConnectorTypes


class CSVConfig(BaseModel):
    overwrite: bool=True
    reporter_type: ConnectorTypes=ConnectorTypes.CSV


import datetime
import json
from collections import defaultdict
from .types.base import UnitTypesMap
from .types import (
    Event,
    Interval,
    Rate
)
from typing import (
    Union, 
    Dict, 
    List,
    Literal,
    Callable,
    Any,
    Optional
)


class MetricStore:

    def __init__(self) -> None:
        self._registered: Dict[
            str,
            Dict[
                str,
                Union[
                    Event,
                    Interval,
                    Rate
                ]
            ]
        ] = defaultdict(dict)

        self._types: Dict[
            Union[
                Literal['event'],
                Literal['interval'],
                Literal['rate']
            ],
            Union[
                Callable[..., Event],
                Callable[..., Interval],
                Callable[..., Rate]
            ]
        ] = {
            'event': lambda kwargs: Event.parse(kwargs),
            'interval': lambda kwargs: Interval.parse(kwargs),
            'rate': lambda kwargs: Rate.parse(kwargs)
        }

        self._units_map = UnitTypesMap()

    def create_load_query(
        self,
        name: str=None,
        kind: Union[
            Literal['event'],
            Literal['interval'],
            Literal['rate']
        ]=None,
        group: str=None,
        timestamp: datetime.datetime=None,
        value: Union[
            int,
            float,
            datetime.datetime,
            str
        ]=None
    ):
        return self._types[kind]({
            'name': name,
            'group': group,
            'timestamp': timestamp,
            'value': value
        })

    def parse(
        self,
        metric_data: Union[
            str,
            bytes,
            Dict[str, Any]
        ]
    ):
        
        if isinstance(metric_data, (str, bytes)):
            metric_data: Dict[str, Any] = json.loads(metric_data)

        timestamp = metric_data.get('timestamp')
        if isinstance(timestamp, str):
            timestamp = datetime.datetime.strptime(
                timestamp,
                '%Y-%m-%dT%H:%M:%S.%f.%z'
            )

            metric_data['timestamp'] = timestamp

        elif isinstance(timestamp, bytes):
            timestamp = datetime.datetime.strptime(
                timestamp.decode(),
                '%Y-%m-%dT%H:%M:%S.%f.%z'
            )

            metric_data['timestamp'] = timestamp

        metric_data['unit'] = self._types.get(
            metric_data.get('unit')
        )
        
        kind: Union[str, bytes] = metric_data.get('kind')
        if kind:
            kind = kind.lower()

        if kind not in self._types:
            raise Exception(f'Err. - Invalid metric kind - {kind} - cannot parse.')

        return self._types[kind](metric_data)

    def record(
        self,
        name: str=None,
        kind: Literal[
            'event',
            'interval',
            'rate'
        ]=None,
        group: str=None,
        timestamp: datetime.datetime=None,
        value: Union[
            int,
            float,
            datetime.datetime,
            str
        ]=None,
        tags: Optional[
            List[
                Dict[str, str]
            ]
        ]=None,
        unit: Literal[
            'nanoseconds',
            'microseconds',
            'milliseconds',
            'seconds',
            'minutes',
            'hours',
            'days',
            'weeks'
        ]=None
    ):
        
        if group is None:
            group = 'default'
        
        metric = self._registered[group].get(name)
        if metric is None:
            metric = self._types[kind]({
                'name': name,
                'group': group,
                'timestamp': timestamp,
                'value': value,
                'tags': tags,
                'unit': self._units_map.get(unit)
            })

            self._registered[group][name] = metric

        else:

            self._registered[group][name].update(
                value,
                tags=tags
            )

        return metric



import uuid
from typing import Union

from metrikflow.logging import MetrikflowLogger
from metrikflow.metrics.types import Event, Interval, Rate
from metrikflow.metrics.types.base import MetricTypes

from .datadog_config import DatadogConfig
from .units import DatadogUnitMap

try:
    # Datadog uses aiosonic
    from aiosonic import HTTPClient, TCPConnector, Timeouts
    from datadog_api_client import AsyncApiClient, Configuration
    from datadog_api_client.v2.api.metrics_api import MetricPayload, MetricsApi
    from datadog_api_client.v2.model.metric_intake_type import MetricIntakeType
    from datadog_api_client.v2.model.metric_point import MetricPoint
    from datadog_api_client.v2.model.metric_resource import MetricResource
    from datadog_api_client.v2.model.metric_series import MetricSeries

    has_connector = True

except Exception:
    HTTPClient=None
    TCPConnector=None
    Timeouts=None
    MetricsApi=None
    MetricPayload=None
    MetricSeries=None
    AsyncApiClient=None
    Configuration=None
    MetricPoint=None
    MetricIntakeType=None
    datadog = None
    has_connector = False


class Datadog:
    has_connector=has_connector

    def __init__(self, config: DatadogConfig) -> None:
        self.datadog_api_key = config.api_key
        self.datadog_app_key = config.app_key
        self.device_name = config.device_name
        self.server_location = config.server_location
        self.priority = config.priority
        self.custom_fields = config.custom_fields or {}

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = MetrikflowLogger()
        self.logger.initialize()
        
        self._config = None
        self._client = None
        self.metrics_api = None
        self.metric_types_map = {
            MetricTypes.INTERVAL: MetricIntakeType.GAUGE,
            MetricTypes.RATE: MetricIntakeType.RATE,
            MetricTypes.EVENT: MetricIntakeType.GAUGE
        }

        self._units_map = DatadogUnitMap()

    async def connect(self):

        self._config = Configuration()
        self._config.api_key["apiKeyAuth"] = self.datadog_api_key
        self._config.api_key["appKeyAuth"] = self.datadog_app_key

        if self.server_location:
            self._config.server_variables["site"] = self.server_location

        self._client = AsyncApiClient(self._config)

        # Datadog's implementation of aiosonic's HTTPClient lacks a lot
        # of configurability, incuding actually being able to set request timeouts
        # so we substitute our own implementation.

        tcp_connection = TCPConnector(timeouts=Timeouts(sock_connect=30))
        self._client.rest_client._client = HTTPClient(tcp_connection)

        self.metrics_api = MetricsApi(self._client)

    async def load(
        self,
        source_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60
    ):
        raise Exception('Err. - Datadog connector is send only.')
        

    async def send(
        self,
        source_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60
    ):

        metric_name = f'{metric.group}.{metric.name}'

        metric_tags = [
            f'{tag_name}:{tag_value}' for tag_name, tag_value in metric.to_tags_dict().items()
        ]

        metric_tags.append(
            f'group:{metric.group}'
        )
        
        series = MetricSeries(
            metric_name, 
            [MetricPoint(
                timestamp=int(metric.timestamp.timestamp()),
                value=metric.value
            )],
            type=self.metric_types_map.get(
                metric.kind,
                MetricIntakeType.GAUGE
            ),
            tags=metric_tags,
            resources=[
                MetricResource(
                    name=self.device_name,
                    type='host'
                )
            ],
            unit=self._units_map.get(metric.unit)
        )

        response = await self.metrics_api.submit_metrics(
            body=MetricPayload(
                series=[series]
            )
        )

        if len(response.errors) > 0:
            for error in response.errors:
                raise Exception(error)
        
    async def close(self):
        pass


from datadog_api_client.v1.model.query_value_widget_definition import QueryValueWidgetDefinition
from datadog_api_client.v1.model.query_value_widget_definition_type import QueryValueWidgetDefinitionType
from datadog_api_client.v1.model.query_value_widget_request import QueryValueWidgetRequest
from datadog_api_client.v1.model.formula_and_function_metric_query_definition import (
    FormulaAndFunctionMetricQueryDefinition,
)
from datadog_api_client.v1.model.formula_and_function_metric_data_source import (
    FormulaAndFunctionMetricDataSource,
)
from datadog_api_client.v1.model.formula_and_function_response_format import (
    FormulaAndFunctionResponseFormat
)
from datadog_api_client.v1.model.widget_formula import WidgetFormula
from datadog_api_client.v1.model.widget_time import WidgetTime
from datadog_api_client.v1.model.widget_style import WidgetStyle
from datadog_api_client.v1.model.widget_text_align import WidgetTextAlign
from metrik.dashboards.datadog.models import (
    WidgetOptions,
    WidgetType
)
from metrik.dashboards.datadog.widgets.widget_types.base.base_widget_factory import BaseWidgetFactory


class QueryValueFactory(BaseWidgetFactory[QueryValueWidgetDefinition]):

    def __init__(self) -> None:
        super().__init__(WidgetType.QUERY_VALUE)

        self._options = [
            'custom_unit',
            'precision'
        ]

    def create(
        self,
        widget_name: str,
        options: WidgetOptions       
    ):
        
        queries = self.queries[widget_name]

        widget_options = options.to_options(self._options)

        query_value = QueryValueWidgetDefinition(
            title=widget_name,
            title_size="16",
            title_align=WidgetTextAlign.LEFT,
            type=QueryValueWidgetDefinitionType.QUERY_VALUE,
            time=WidgetTime(
                live_span=options.to_datadog_live_span()
            ),
            autoscale=True,
            requests=[
                QueryValueWidgetRequest(
                    formulas=[
                        WidgetFormula(
                            formula=' + '.join([
                                query.to_query_name() for query in queries
                            ])
                        )
                    ],
                    queries=[
                        FormulaAndFunctionMetricQueryDefinition(
                            data_source=FormulaAndFunctionMetricDataSource.METRICS,
                            name=query.to_query_name(),
                            aggregator=query.formula_type,
                            query=query.to_query()
                        ) if query.aggregation_field else FormulaAndFunctionMetricQueryDefinition(
                            data_source=FormulaAndFunctionMetricDataSource.METRICS,
                            name=query.to_query_name(),
                            query=query.to_query()
                        )  for query in queries
                    ],
                    response_format=FormulaAndFunctionResponseFormat.SCALAR,
                )
            ],
            **widget_options
        )

        self._widgets[widget_name] = query_value

        return query_value


import asyncio
import functools
import os
import uuid
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Dict, List, Union

from metrikflow.connectors.common.signals import add_signal_handler, handle_loop_stop
from metrikflow.logging import MetrikflowLogger
from metrikflow.metrics import MetricStore
from metrikflow.metrics.types import Event, Interval, Rate

from .snowflake_config import SnowflakeConfig

try:
    import sqlalchemy
    from snowflake.sqlalchemy import URL
    from sqlalchemy import create_engine
    from sqlalchemy.schema import CreateTable
    
    has_connector = True

except Exception:
    snowflake = None
    has_connector = False

class Snowflake:
    has_connector=has_connector

    def __init__(self, config: SnowflakeConfig) -> None:
        self.username = config.username
        self.password = config.password
        self.organization_id = config.organization_id
        self.account_id = config.account_id
        self.private_key = config.private_key
        self.warehouse = config.warehouse
        self.database = config.database
        self.schema = config.database_schema

        self.connect_timeout = config.connect_timeout
        
        self.metadata = sqlalchemy.MetaData()
        self._executor = ThreadPoolExecutor(
            max_workers=os.cpu_count()
        )

        self._engine = None
        self._connection = None

        self._loop = asyncio.get_event_loop()

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = MetrikflowLogger()
        self.logger.initialize()
        self._store = MetricStore()

    async def connect(self):

        try:

            add_signal_handler(
                self._loop,
                handle_loop_stop,
                self._executor,
                self._loop
            )
            
            self._engine = await self._loop.run_in_executor(
                self._executor,
                create_engine,
                URL(
                    user=self.username,
                    password=self.password,
                    account=self.account_id,
                    warehouse=self.warehouse,
                    database=self.database,
                    schema=self.schema
                )
            
            )

            self._connection = await asyncio.wait_for(
                self._loop.run_in_executor(
                    self._executor,
                    self._engine.connect
                ),
                timeout=self.connect_timeout
            )

        except asyncio.TimeoutError:
            raise Exception('Err. - Connection to Snowflake timed out - check your account id, username, and password.')
        
    async def load(
        self,
        source_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60  
    ):
        table = sqlalchemy.Table(
            source_name,
            self.metadata,
            sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
            sqlalchemy.Column('metric_id', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('kind', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('group', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('timestamp', sqlalchemy.DateTime()),
            sqlalchemy.Column('value', sqlalchemy.Float())
        )

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._connection.execute,
                CreateTable(
                    table, 
                    if_not_exists=True
                )
            )
        )

        rows: List[Any] = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._connection.execute,
                table.select().where(
                    table.c.name == metric.name
                ).order_by(
                    table.c.timestamp.desc()
                ).limit(1)
            )
        )

        metrics_data: List[Dict[str, Any]] = []
        for row in rows:
            metrics_data.append({
                'metric_id': row.metric_id,
                'name': row.name,
                'kind': row.kind,
                'group': row.group,
                'timestamp': row.timestamp,
                'value': row.value
            })

        return self._store.parse(
            metrics_data[-1]
        )

        
    async def send(
        self,
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60   
    ):
        table = sqlalchemy.Table(
            destination_name,
            self.metadata,
            sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
            sqlalchemy.Column('metric_id', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('kind', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('group', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('timestamp', sqlalchemy.DateTime()),
            sqlalchemy.Column('value', sqlalchemy.Float())
        )

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._connection.execute,
                CreateTable(
                    table, 
                    if_not_exists=True
                )
            )
        )

        metric_data = metric.dict()
        metric_data['kind'] = metric.kind.value

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._connection.execute,
                table.insert().values(**metric_data)
            )
        )

    async def close(self):

        await self._loop.run_in_executor(
            self._executor,
            self._connection.close
        )

        self._executor.shutdown(wait=False, cancel_futures=True)



from collections import defaultdict
from metrik.dashboards.datadog.layouts.size_type import SizeType
from metrik.dashboards.datadog.models import (
    Query,
    WidgetType
)
from typing import Literal, List, Dict, TypeVar, Generic

T = TypeVar('T')


class BaseWidgetFactory(Generic[T]):

    def __init__(
        self,
        kind: WidgetType
    ) -> None:
        self.sizes: Dict[str, SizeType] = {}
        self.kind = kind
        self.queries: Dict[str, List[Query]] = defaultdict(list)
        self._widgets: Dict[str, T] = {}

    @property
    def kind_name(self):
        return self.kind.value.lower()
    
    def create_widget(
        self,
        widget_name: str,
        size: SizeType
    ):
        self.sizes[widget_name] = size
    
    def add_metric(
        self,
        widget_name: str,
        query: Query
    ):
        self.queries[widget_name].append(query)


from __future__ import annotations
import datetime
import uuid
from typing import (
    Any, 
    Dict, 
    Union, 
    Optional, 
    List
)

from pydantic import StrictFloat, StrictInt, StrictStr
from metrik.metrics.types.base import (
    Metric,
    MetricTypes,
    UnitType
)


class Event(Metric):

    def __init__(
        self,
        metric_id: str=None,
        name: str=None,
        group: str=None,
        timestamp: datetime.datetime=None,
        value: Union[
            int,
            float,
            datetime.datetime,
            str
        ]=None,
        tags: Optional[
            List[
                Dict[str, str]
            ]
        ]=None,
        unit: Optional[UnitType]=None
    ):
        
        if metric_id is None:
            metric_id = str(uuid.uuid4())
        
        if value is None:
            value = 1

        if tags is None:
            tags = []

        super().__init__(
            name=name,
            kind=MetricTypes.EVENT,
            group=group,
            timestamp=timestamp,
            value=value,
            tags=tags,
            unit=unit
        )

    @classmethod
    def parse(self, data: Dict[str, Any]):
        return Event(
            metric_id=data.get('metric_id'),
            name=data.get('name'),
            group=data.get('group'),
            timestamp=data.get('timestamp'),
            value=data.get('value'),
            tags=data.get('tags'),
            unit=data.get('unit')
        )
    
    def update(
        self,
        value: Union[
            int,
            float,
            datetime.datetime,
            str
        ],
        tags: Optional[
            List[
                Dict[str, str]
            ]
        ]=None
    ):
        if value is None:
            value = 1

        update_tags = []
        if tags:
            update_tags = tags
            update_tags.extend(self.tags)

        return Event(
            metric_id=self.metric_id,
            name=self.name,
            group=self.group,
            timestamp=datetime.datetime.now(
                datetime.timezone.utc
            ),
            value=value,
            tags=update_tags,
            unit=self.unit
        )


from datadog_api_client.v1.model.table_widget_definition import TableWidgetDefinition
from datadog_api_client.v1.model.table_widget_definition_type import TableWidgetDefinitionType
from datadog_api_client.v1.model.table_widget_cell_display_mode import TableWidgetCellDisplayMode
from datadog_api_client.v1.model.table_widget_has_search_bar import TableWidgetHasSearchBar
from datadog_api_client.v1.model.table_widget_request import TableWidgetRequest
from datadog_api_client.v1.model.widget_text_align import WidgetTextAlign
from datadog_api_client.v1.model.widget_formula import WidgetFormula
from datadog_api_client.v1.model.widget_formula_limit import WidgetFormulaLimit
from datadog_api_client.v1.model.widget_style import WidgetStyle
from datadog_api_client.v1.model.widget_time import WidgetTime
from datadog_api_client.v1.model.formula_and_function_metric_query_definition import (
    FormulaAndFunctionMetricQueryDefinition,
)
from datadog_api_client.v1.model.formula_and_function_metric_data_source import (
    FormulaAndFunctionMetricDataSource,
)
from datadog_api_client.v1.model.formula_and_function_response_format import (
    FormulaAndFunctionResponseFormat
)
from metrik.dashboards.datadog.models import (
    WidgetOptions,
    WidgetType
)
from metrik.dashboards.datadog.widgets.widget_types.base.base_widget_factory import BaseWidgetFactory


class TableFactory(BaseWidgetFactory[TableWidgetDefinition]):

    def __init__(self) -> None:
        super().__init__(WidgetType.TABLE)

    def create(
        self,
        widget_name: str,
        options: WidgetOptions
    ):
        
        queries = self.queries[widget_name]
        
        table = TableWidgetDefinition(
            title=widget_name,
            title_size='16',
            title_align=WidgetTextAlign.LEFT,
            type=TableWidgetDefinitionType.QUERY_TABLE,
            has_search_bar=options.to_search_bar_type(),
            time=WidgetTime(
                live_span=options.to_datadog_live_span()
            ),
            requests=[
                TableWidgetRequest(
                    formulas=[
                        WidgetFormula(
                            formula=' + '.join([
                                query.to_query_name() for query in queries
                            ]),
                            limit=WidgetFormulaLimit(
                                count=options.limit,
                                order=options.to_datadog_sort_order()
                            ),
                            cell_display_mode=options.to_datadog_cell_display_mode()
                        )
                    ],
                    queries=[
                        FormulaAndFunctionMetricQueryDefinition(
                            data_source=FormulaAndFunctionMetricDataSource.METRICS,
                            name=query.to_query_name(),
                            aggregator=query.formula_type,
                            query=query.to_query()
                        ) if query.aggregation_field else FormulaAndFunctionMetricQueryDefinition(
                            data_source=FormulaAndFunctionMetricDataSource.METRICS,
                            name=query.to_query_name(),
                            query=query.to_query()
                        ) for query in queries
                    ],
                    response_format=FormulaAndFunctionResponseFormat.SCALAR
                )
            ]
        )

        self._widgets[widget_name] = table

        return table


from enum import Enum


class WidgetType(Enum):
    CHANGE='CHANGE'
    HISTOGRAM='HISTOGRAM'
    QUERY_VALUE='QUERY_VALUE'
    SCATTER_PLOT='SCATTER_PLOT'
    TABLE='TABLE'
    TIMESERIES='TIMESERIES'


import os

import click

from metrikflow.cli.commands import query_metric


@click.command(help='Query for Metrics via various connectors')
@click.option(
    '--metric-id',
    help='Metric ID of the metric to query for.'
)
@click.option(
    '--name',
    help='Name of the metric to query for.'
)
@click.option(
    '--kind',
    help='Kind of metric to query for (event, interval, rate, etc.).'
)
@click.option(
    '--group',
    default='default',
    help='Group of metric to query for.'
)
@click.option(
    '--newer-than',
    default='30d',
    help='Only fetch metrics newer than the specified interval.'
)
@click.option(
    '--older-than',
    help='Only fetch metrics older than the specified interval.'
)
@click.option(
    '--search',
    help='A string specified as <format>:<location/path/uri> denoting where to load existing metric from.'
)
@click.option(
    '--config',
    show_default=True,
    default=f'{os.getcwd()}/.metrikflow.json',
    help='Path to existing .metrikflow.json.'
)
@click.option(
    '--timeout',
    show_default=True,
    default=60,
    type=float,
    help='Timeout (in seconds) to send metric for storage via selected Connector format.'
)
def query(
    metric_id: str,
    name: str,
    kind: str,
    group: str,
    newer_than: str,
    older_than: str,
    search: str,
    config: str,
    timeout: float
):
    query_metric(
        metric_id=metric_id,
        name=name,
        kind=kind,
        group=group,
        newer_than=newer_than,
        older_than=older_than,
        search=search,
        config=config,
        timeout=timeout
    )




from enum import Enum


class UnitType(Enum):
    NANOSECONDS='NANOSECONDS'
    MICROSECONDS='MICROSECONDS'
    MILLISECONDS='MILLISECONDS'
    SECONDS='SECONDS'
    MINUTES='MINUTES'
    HOURS='HOURS'
    DAYS='DAYS'
    WEEKS='WEEKS'


import os

import click

from metrikflow.cli.commands import record_metric


@click.group(help='Generate new metrics or update existing ones.')
def record():
    pass

@record.command(
    help='Record a new instance of an Event metric.'
)
@click.argument('name')
@click.argument(
    'value',
    default=1,
    type=float
)
@click.option(
    '--group',
    default='default',
    help='Group of newly created metric.'
)
@click.option(
    '--load',
    help='A string specified as <format>:<location/path/uri> denoting where to load existing metric from.'
)
@click.option(
    '--send',
    help='A string specified as <format>:<location/path/uri> denoting where to load existing metric from.'
)
@click.option(
    '--config',
    show_default=True,
    default=f'{os.getcwd()}/.metrikflow.json',
    help='Path to existing .metrikflow.json.'
)
@click.option(
    '--timeout',
    show_default=True,
    default=60,
    type=float,
    help='Timeout (in seconds) to send metric for storage via selected Connector format.'
)
@click.option(
    '--tags',
    help="Comma-delimited list of <key>:<value> pairs to add to a metric's tags."
)
def event(
    name: str,
    group: str,
    value: float,
    load: str,
    send: str,
    config: str,
    timeout: float,
    tags: str
):
    record_metric(
        name=name,
        kind='event',
        group=group,
        value=value,
        load=load,
        send=send,
        config_path=config,
        timeout=timeout,
        tags=tags
    )

@record.command(
    help='Record a new instance of an Interval metric.'
)
@click.argument('name')
@click.option(
    '--group',
    default='default',
    help='Group of newly created metric.'
)
@click.option(
    '--load',
    help='A string specified as <format>:<location/path/uri> denoting where to load existing metric from.'
)
@click.option(
    '--send',
    help='A string specified as <format>:<location/path/uri> denoting where to load existing metric from.'
)
@click.option(
    '--config',
    show_default=True,
    default=f'{os.getcwd()}/.metrikflow.json',
    help='Path to existing .metrikflow.json.'
)
@click.option(
    '--timeout',
    show_default=True,
    default=60,
    type=float,
    help='Timeout (in seconds) to send metric for storage via selected Connector format.'
)
@click.option(
    '--tags',
    help="Comma-delimited list of <key>:<value> pairs to add to a metric's tags."
)
def interval(
    name: str,
    group: str,
    load: str,
    send: str,
    config: str,
    timeout: float,
    tags: str
):
    
    record_metric(
        name=name,
        kind='interval',
        group=group,
        load=load,
        send=send,
        config_path=config,
        timeout=timeout,
        tags=tags
    )

@record.command(
    help='Record a new instance of a Rate metric.'
)
@click.argument('name')
@click.argument(
    'value', 
    default=1,
    type=float
)
@click.option(
    '--group',
    default='default',
    help='Group of newly created metric.'
)
@click.option(
    '--load',
    help='A string specified as <format>:<location/path/uri> denoting where to load existing metric from.'
)
@click.option(
    '--send',
    help='A string specified as <format>:<location/path/uri> denoting where to load existing metric from.'
)
@click.option(
    '--config',
    show_default=True,
    default=f'{os.getcwd()}/.metrikflow.json',
    help='Path to existing .metrikflow.json.'
)
@click.option(
    '--timeout',
    show_default=True,
    default=60,
    type=float,
    help='Timeout (in seconds) to send metric for storage via selected Connector format.'
)
@click.option(
    '--tags',
    help="Comma-delimited list of <key>:<value> pairs to add to a metric's tags."
)
def rate(
    name: str,
    group: str,
    value: float,
    load: str,
    send: str,
    config: str,
    timeout: float,
    tags: str
):
    
    record_metric(
        name=name,
        kind='rate',
        group=group,
        value=value,
        load=load,
        send=send,
        config_path=config,
        timeout=timeout,
        tags=tags
    )


from typing import Optional
from metrik.connectors.common.types import ConnectorTypes
from pydantic import BaseModel


class RedisConfig(BaseModel):
    host: str='localhost:6379'
    username: Optional[str]
    password: Optional[str]
    database: int=0
    secure: bool=False
    reporter_type: ConnectorTypes=ConnectorTypes.Redis


from pydantic import (
    BaseModel,
    StrictInt,
    StrictFloat
)
from typing import Union


class RateTimingSet(BaseModel):
    current: StrictFloat=0.0
    last_measured: Union[StrictInt, StrictFloat]
    last_time: StrictFloat


from collections import defaultdict
from typing import Dict
from .datadog_dashboard import DatadogDashboard
from .datadog_config import DatadogConfig


class DatadogDashboardFactory:

    def __init__(
        self,
        config: DatadogConfig
    ) -> None:
        self._config = config
        self._dashboards: Dict[
            str,
            DatadogDashboard
        ] = defaultdict(
            lambda: DatadogDashboard(self._config)
        )

    def __getitem__(self, dashboard_name: str):
        return self._dashboards[dashboard_name]
    
    def get(self, dashboard_name: str):
        return self._dashboards[dashboard_name]


from __future__ import annotations

import asyncio
import datetime
import functools
import json
import os
import re
import uuid
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Dict, List, TextIO, Union

from metrikflow.connectors.common.signals import (
    add_signal_handler,
    handle_file_stop,
    handle_loop_stop,
)
from metrikflow.logging import MetrikflowLogger
from metrikflow.metrics import MetricStore
from metrikflow.metrics.types import Event, Interval, Rate

from .json_config import JSONConfig

has_connector = True

class JSON:
    has_connector=has_connector

    def __init__(self, config: JSONConfig) -> None:

        self._executor = ThreadPoolExecutor(
            max_workers=os.cpu_count()
        )
        self._loop: asyncio.AbstractEventLoop = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = MetrikflowLogger()
        self.logger.initialize()

        self.metrics_file: TextIO = None

        self.write_mode = 'w' if config.overwrite else 'a'
        self.pattern = re.compile("_copy[0-9]+")
        self._store = MetricStore()

    async def connect(self):
        self._loop = asyncio._get_running_loop()

        add_signal_handler(
            self._loop,
            handle_loop_stop,
            self._executor,
            self._loop
        )

    
    async def load(
        self,
        source_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60       
    ):
        absolute_path = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                os.path.abspath,
                source_name
            )
        )

        file_exists = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                os.path.exists,
                absolute_path
            )
        )

        if file_exists is False:
            return None

        self.metrics_file = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                open,
                absolute_path,
                'r'
            )
        )

        add_signal_handler(
            self._loop,
            handle_file_stop,
            self.metrics_file
        )

        metrics_data: Dict[
            str, 
            List[Dict[str, Any]]
        ] = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                json.load,
                self.metrics_file
            )
        )

        metrics: List[
            Dict[str, Any]
        ] = list(
            sorted(
                metrics_data.get(metric.name, []),
                key=lambda metric: datetime.datetime.strptime(
                    metric.get('timestamp'),
                    '%Y-%m-%dT%H:%M:%S.%f.%z'
                )
            )
        )

        if len(metrics) < 1:
            return None

        return self._store.parse(
            metrics[-1]
        )

    async def send(
        self,
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60
    ):
        absolute_path = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                os.path.abspath,
                destination_name
            )
        )

        file_exists = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                os.path.exists,
                absolute_path
            )
        )

        data = metric.to_data_dict()

        add_signal_handler(
            self._loop,
            handle_file_stop,
            self.metrics_file
        )


        if file_exists:

            self.metrics_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    absolute_path,
                    'r'
                )
            )
  
            metrics_data: Dict[
                str, 
                List[Dict[str, Any]]
            ] = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    json.load,
                    self.metrics_file
                )
            )

            metrics = metrics_data.get(metric.name, [])
            metrics.append(data)
            metrics_data[metric.name] = metrics

            await self._loop.run_in_executor(
                self._executor,
                self.metrics_file.close
            )

            self.metrics_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    absolute_path,
                    'w'
                )
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    json.dump,
                    metrics_data,
                    self.metrics_file,
                    indent=4
                )
            )
            
        else:

            self.metrics_file = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    open,
                    absolute_path,
                    'w+'
                )
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    json.dump,
                    {
                        metric.name: [
                            data
                        ]
                    },
                    self.metrics_file,
                    indent=4
                )
            )

    async def close(self):

        if self.metrics_file:
            await self._loop.run_in_executor(
                self._executor,
                self.metrics_file.close
            )

        self._executor.shutdown(wait=False, cancel_futures=True)


from typing import Dict, Literal, Union
from .widget_type import WidgetType


class WidgetTypeMap:

    def __init__(self) -> None:
        self._types: Dict[
            Literal[
                'change',
                'histogram',
                'query_value',
                'scatter_plot',
                'table',
                'timeseries'
            ],
            WidgetType
        ] = {
            'change': WidgetType.CHANGE,
            'histogram': WidgetType.HISTOGRAM,
            'query_value': WidgetType.QUERY_VALUE,
            'scatter_plot': WidgetType.SCATTER_PLOT,
            'table': WidgetType.TABLE,
            'timeseries': WidgetType.TIMESERIES
        }

    def __iter__(self):
        for widget_type in self._types.values():
            yield widget_type

    def __getitem__(
        self,
        widget_type: Literal['timeseries']
    ):
        return self._types.get(
            widget_type,
            WidgetType.TIMESERIES
        )
    
    def get(
        self,
        widget_type: Literal[
            'change',
            'histogram',
            'query_value',
            'scatter_plot',
            'table',
            'timeseries'
        ]
    ):
        return self._types.get(widget_type)


import re
from pydantic import (
    BaseModel,
    StrictStr
)
from typing import Union, Literal, Optional, Any

class Query(BaseModel):
    aggregator: Union[
        Literal['avg'],
        Literal['sum'],
        Literal['min'],
        Literal['max']
    ]='avg'
    metric: StrictStr
    aggregation_field: Optional[StrictStr]=None
    formula_type: Optional[Any]=None

    class Config:
        arbitrary_types_allowed=True

    def to_query(self):

        metric_name = self.metric.replace('-', '_')
        query_string = [
            f'{self.aggregator}:{metric_name}', 
            '{*}'
        ]
     

        if self.aggregation_field:
            query_string.extend([
                ' by {',
                self.aggregation_field,
                '}'
            ])

        return ''.join(query_string)
    
    def to_query_name(self):
        return re.sub(r"[&.-]", "_", self.metric)


from pydantic import BaseModel
from metrik.connectors.common.types import ConnectorTypes
from typing import Optional


class AWSLambdaConfig(BaseModel):
    aws_access_key_id: str
    aws_secret_access_key: str
    region_name: str
    reporter_type: ConnectorTypes=ConnectorTypes.AWSLambda


from typing import (
    Dict,
    Callable
)
from .datadog import (
    DatadogDashboardFactory,
    DatadogConfig,
    DatadogDashboard
)
from .datadog.models.widget_options import WidgetOptions
from typing import Any
from .common import DashboardType


class Dashboard:

    def __init__(
        self
    ) -> None:
        self._factories: Dict[
            DashboardType.DATADOG,
            DatadogDashboardFactory
        ] = {}

        self._dashboards: Dict[
            str,
            DatadogDashboard
        ] = {}

    def __getitem__(
        self,
        dashboard_name: str
    ):
        return self._dashboards.get(dashboard_name)
    
    def get(
        self,
        dashboard_name: str
    ):
        return self._dashboards.get(dashboard_name)

    async def create_datadog_dashboard(
        self,
        config: Dict[
            str,
            Any
        ],
        dashboard_name: str
    ):
        if self._factories.get(DashboardType.DATADOG) is None:
            self._factories[DashboardType.DATADOG] = DatadogDashboardFactory(
                DatadogConfig(**config)
            )

        new_dashboard = self._factories[DashboardType.DATADOG].get(dashboard_name)

        await new_dashboard.connect()

        self._dashboards[dashboard_name] = new_dashboard

        return new_dashboard
    
    async def publish(
        self,
        dashboard_name: str,
        description: str,
        options: WidgetOptions=None
    ):
        return await self._dashboards[dashboard_name].create_dashboard(
            dashboard_name,
            description,
            options=options
        )


from metrik.connectors.common.types import ConnectorTypes
from pydantic import BaseModel


class PostgresConfig(BaseModel):
    host: str='localhost'
    database: str
    username: str
    password: str
    reporter_type: ConnectorTypes=ConnectorTypes.Postgres

    class Config:
        arbitrary_types_allowed = True


# Datadog uses aiosonic
from aiosonic import HTTPClient, TCPConnector, Timeouts
from datadog_api_client import (
    AsyncApiClient, 
    Configuration
)
from datadog_api_client.v1.api.dashboards_api import DashboardsApi
from datadog_api_client.v1.model.dashboard import Dashboard
from datadog_api_client.v1.model.dashboard_summary import DashboardSummary
from datadog_api_client.v1.model.dashboard_layout_type import DashboardLayoutType
from datadog_api_client.v1.model.widget import Widget
from metrik.metrics.types import (
    Event,
    Interval,
    Rate
)
from typing import List, Union, Literal, Optional, Dict, Any
from .datadog_config import DatadogConfig
from .models import (
    DatadogDashboardSummary,
    WidgetMetric, 
    WidgetOptions
)
from .widgets import WidgetsFactory

class DatadogDashboard:

    def __init__(self, config: DatadogConfig) -> None:
        self.datadog_api_key = config.api_key
        self.datadog_app_key = config.app_key
        self.device_name = config.device_name
        self.server_location = config.server_location
        self.priority = config.priority

        self._widgets: List[Widget] = []
        self._config: Union[Configuration, None] = None
        self._client: Union[AsyncApiClient, None] = None
        self._dashboards_api: Union[DashboardsApi, None] = None
        self._widgets_factory = WidgetsFactory()
        self._summaries: Dict[str, DatadogDashboardSummary] = {}
        self._summary_titles: Dict[str, str] = {}
        self._title_id_map: Dict[str, str] = {}

    @property
    def ids(self):
        return list(self._summary_titles.keys())

    @property
    def names(self):
        return list(self._summary_titles.values())
    
    def get_id_by_title(
        self,
        title: str
    ):
        return self._title_id_map.get(title)

    async def connect(self):
        self._config = Configuration()

        self._config.api_key["apiKeyAuth"] = self.datadog_api_key
        self._config.api_key["appKeyAuth"] = self.datadog_app_key

        if self.server_location:
            self._config.server_variables["site"] = self.server_location

        self._client = AsyncApiClient(self._config)

        # Datadog's implementation of aiosonic's HTTPClient lacks a lot
        # of configurability, incuding actually being able to set request timeouts
        # so we substitute our own implementation.

        tcp_connection = TCPConnector(timeouts=Timeouts(sock_connect=30))
        self._client.rest_client._client = HTTPClient(tcp_connection)

        self._dashboards_api = DashboardsApi(self._client)

    def add_metric(
        self,
        widget: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        kind: Literal[
            'change',
            'histogram',
            'query_value',
            'scatter_plot',
            'table',
            'timeseries'
        ]='timeseries',
        size: Union[
            Literal['extra_small'],
            Literal['small'],
            Literal['medium'],
            Literal['large'],
            Literal['extra_large']
        ]='medium',
        aggregator: Union[
            Literal['average'],
            Literal['total'],
            Literal['maximum'],
            Literal['minimum']
        ]='average',
        aggregator_field: Optional[str]=None,
        options: Optional[WidgetOptions]=None
    ):
        self._widgets_factory.add_metric(
            WidgetMetric(
                widget_name=widget,
                widget_type=kind,
                metric=metric,
                widget_size=size,
                widget_aggregator=aggregator,
                widget_aggregator_field=aggregator_field,
                widget_options=options
            )
        )

    async def create_dashboard(
        self,
        title: str,
        description: str,
        options: Optional[WidgetOptions]=None
    ) -> Dashboard:
        widgets = self._widgets_factory.to_widgets(options=options)
        dashboard = Dashboard(
            title=title,
            description=description,
            widgets=widgets,
            layout_type=DashboardLayoutType.ORDERED
        )

        response: Dashboard = await self._dashboards_api.create_dashboard(dashboard)
        self._store_dashboard_as_summary(response)

        return response
    
    async def get_dashboard(
        self,
        dashboard_id: str
    ) -> Dashboard:
        
        response: Dashboard = await self._dashboards_api.get_dashboard(dashboard_id)
        self._store_dashboard_as_summary(response)

        return response
    
    async def list_api_dashboards(
        self,
        limit: int=50,
        offset: int=0,
        hide_shared: bool=False,
        hide_deleted: bool=False
    ):
        response: DashboardSummary = await self._dashboards_api.list_dashboards(
            filter_shared=hide_shared,
            filter_deleted=hide_deleted,
            count=limit,
            start=offset
        )

        dashboards = [
            DatadogDashboardSummary(
                **dashboard
            ) for dashboard in response.to_dict().get('dashboards', [])
        ]

        for dashboard in dashboards:
            self._summaries[dashboard.id] = dashboard
            self._summary_titles[dashboard.id] = dashboard.title
            self._title_id_map[dashboard.title] = dashboard.id

        return dashboards
    
    async def list_all_api_dashboards(
        self,
        limit: int=100
    ):
        
        discovered = await self.list_api_dashboards(
            limit=limit
        )

        discovered_count = len(discovered)
        offset = discovered_count
        
        while discovered_count == limit:
            discovered = await self.list_api_dashboards(
                offset=offset
            )

            discovered_count = len(discovered)
            offset += discovered_count

        return list(self._summaries.values())
    
    async def update_dashboard(
        self,
        dashboard_id: str,
        title: str,
        description: str,
        options: Optional[WidgetOptions]=None
    ):
        
        widgets = self._widgets_factory.to_widgets(options=options)
        dashboard = Dashboard(
            title=title,
            description=description,
            widgets=widgets,
            layout_type=DashboardLayoutType.ORDERED
        )

        return await self._dashboards_api.update_dashboard(
            dashboard_id,
            dashboard
        )
    
    async def delete_dashboard(
        self,
        dashboard_id: str
    ):
        return await self._dashboards_api.delete_dashboard(
            dashboard_id
        )

    def _store_dashboard_as_summary(
        self,
        dashboard: Dashboard
    ) -> DashboardSummary:
        dashboard_summary_data = {
            key: value for key, value in dashboard.to_dict().items() if key in DatadogDashboardSummary.fields()
        }

        dashboard_summary = DatadogDashboardSummary(**dashboard_summary_data)
        self._summaries[dashboard_summary.id] = dashboard_summary
        self._summary_titles[dashboard_summary.id] = dashboard_summary.title
        self._title_id_map[dashboard.title] = dashboard_summary.id

        return dashboard_summary


from pydantic import BaseModel
from metrik.connectors.common.types import ConnectorTypes



class MySQLConfig(BaseModel):
    host: str='127.0.0.1'
    database: str
    username: str
    password: str
    reporter_type: ConnectorTypes=ConnectorTypes.MySQL

    class Config:
        arbitrary_types_allowed = True


from typing import Literal, Dict
from .unit_type import UnitType


class UnitTypesMap:

    def __init__(self) -> None:
        self._types: Dict[
            Literal[
                'nanoseconds',
                'microseconds',
                'milliseconds',
                'seconds',
                'minutes',
                'hours',
                'days',
                'weeks'
            ],
            UnitType
        ] = {
            'nanoseconds': UnitType.NANOSECONDS,
            'microseconds': UnitType.MICROSECONDS,
            'milliseconds': UnitType.MILLISECONDS,
            'seconds': UnitType.SECONDS,
            'minutes': UnitType.MINUTES,
            'hours': UnitType.HOURS,
            'days': UnitType.DAYS,
            'weeks': UnitType.WEEKS
        }

    def __getitem__(
        self,
        unit_type: Literal[
            'nanoseconds',
            'microseconds',
            'milliseconds',
            'seconds',
            'minutes',
            'hours',
            'days',
            'weeks'
        ]
    ):
        return self._types.get(unit_type)

    def get(
        self,
        unit_type: Literal[
            'nanoseconds',
            'microseconds',
            'milliseconds',
            'seconds',
            'minutes',
            'hours',
            'days',
            'weeks'
        ] 
    ):
        return self._types.get(unit_type)


from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictBool,
    validator
)
from datadog_api_client.v1.model.table_widget_cell_display_mode import TableWidgetCellDisplayMode
from datadog_api_client.v1.model.widget_live_span import WidgetLiveSpan
from datadog_api_client.v1.model.query_sort_order import QuerySortOrder
from datadog_api_client.v1.model.widget_order_by import WidgetOrderBy
from datadog_api_client.v1.model.table_widget_has_search_bar import TableWidgetHasSearchBar
from typing import (
    Literal,
    Union,
    Optional,
    List,
    Dict
)


class WidgetOptions(BaseModel):
    cell_display_mode: Union[
        Literal['bar'],
        Literal['number']
    ]='number'
    comparison_window: Union[
        Literal['day_before'],
        Literal['hour_before'],
        Literal['month_before'],
        Literal['week_before']
    ]='hour_before'
    comparison_type: Union[
        Literal['absolute'],
        Literal['relative']
    ]='absolute'
    limit: StrictInt=10
    order: Union[
        Literal['asc'],
        Literal['desc']
    ]='desc'
    order_by: Literal[
        'change',
        'name',
        'past',
        'present'
    ]='change',
    show_increase_as_positive: StrictBool=True
    custom_unit: Optional[StrictStr]=None
    precision: StrictInt=2
    search_bar_type: Literal[
        'always',
        'never',
        'auto'
    ]='auto'
    window: Literal[
        '1m',
        '5m',
        '10m',
        '15m',
        '30m',
        '1h',
        '4h',
        '1d',
        '2d',
        '1w',
        '1mo',
        '3mo',
        '6mo',
        '1y',
        'alert'
    ]='1d'
    x_axis_name: Optional[StrictStr]=None
    y_axis_name: Optional[StrictStr]=None
    x_axis_metrics: List[StrictStr]=[]
    y_axis_metrics: List[StrictStr]=[]
    display_current_values: StrictBool=True

    @validator('x_axis_name')
    def validate_x_axis_metrics(cls, v):

        if cls.x_axis_name is not None:
            assert len(cls.x_axis_metrics) > 0, "At least one metric must be specified as part of the X axis if an X axis is specified"

    @validator('y_axis_name')
    def validate_y_axis_metrics(cls, v):

        if cls.y_axis_name is not None:
            assert len(cls.y_axis_metrics) > 0, "At least one metric must be specified as part of the Y axis if an Y axis is specified"

    def to_datadog_sort_order(self):
        return QuerySortOrder.ASC if self.order == 'asc' else QuerySortOrder.DESC
    
    def to_datadog_live_span(self):

        live_spans: Dict[
            Literal[
                '1m',
                '5m',
                '10m',
                '15m',
                '30m',
                '1h',
                '4h',
                '1d',
                '2d',
                '1w',
                '1mo',
                '3mo',
                '6mo',
                '1y',
                'week_to_date',
                'month_to_date',
                'alert'
            ]
        ] = {
            '1m': WidgetLiveSpan.PAST_ONE_MINUTE,
            '5m': WidgetLiveSpan.PAST_FIVE_MINUTES,
            '10m': WidgetLiveSpan.PAST_TEN_MINUTES,
            '15m': WidgetLiveSpan.PAST_FIFTEEN_MINUTES,
            '30m': WidgetLiveSpan.PAST_THIRTY_MINUTES,
            '1h': WidgetLiveSpan.PAST_ONE_HOUR,
            '4h': WidgetLiveSpan.PAST_FOUR_HOURS,
            '1d': WidgetLiveSpan.PAST_ONE_DAY,
            '2d': WidgetLiveSpan.PAST_TWO_DAYS,
            '1w': WidgetLiveSpan.PAST_ONE_WEEK,
            '1mo': WidgetLiveSpan.PAST_ONE_MONTH,
            '3mo': WidgetLiveSpan.PAST_THREE_MONTHS,
            '6mo': WidgetLiveSpan.PAST_SIX_MONTHS,
            '1y': WidgetLiveSpan.PAST_ONE_YEAR,
            'alert': WidgetLiveSpan.ALERT

        }

        return live_spans.get(
            self.window,
            WidgetLiveSpan.PAST_ONE_DAY
        )
    
    def to_datadog_cell_display_mode(self):
        return TableWidgetCellDisplayMode.BAR if self.cell_display_mode == 'bar' else TableWidgetCellDisplayMode.NUMBER
    
    def to_widget_order_by(self):
        order_types: Dict[
            str,
            WidgetOrderBy
        ] = {
            'change': WidgetOrderBy.CHANGE,
            'name': WidgetOrderBy.NAME,
            'past': WidgetOrderBy.PAST,
            'present': WidgetOrderBy.PRESENT
        }

        return order_types.get(
            self.order_by,
            WidgetOrderBy.CHANGE
        )
    
    def to_search_bar_type(self):
        search_bar_types: Dict[
            str,
            TableWidgetHasSearchBar
        ] = {
            'always': TableWidgetHasSearchBar.ALWAYS,
            'never': TableWidgetHasSearchBar.NEVER,
            'auto': TableWidgetHasSearchBar.AUTO
        }

        return search_bar_types.get(
            self.search_bar_type,
            TableWidgetHasSearchBar.AUTO
        )
    
    def to_options(
        self,
        widget_supported_options: List[str]
    ):
        return {
            option_name: option_value for option_name, option_value in self.dict().items() if option_name in widget_supported_options and option_value is not None
        }


import datetime
from pydantic import (
    BaseModel,
    StrictStr
)
from typing import Union, Optional, Literal
from .types.base import MetricTypes


class MetricQuery(BaseModel):
    name: Optional[StrictStr]=None
    kind: Optional[
        Union[
            Literal['event'],
            Literal['interval'],
            Literal['rate']
        ]
    ]=None,
    group: Optional[StrictStr]=None
    newer_than: Optional[datetime.datetime]=None
    older_than: Optional[datetime.datetime]=None




import asyncio
import functools
import json
import os
import uuid
from concurrent.futures import ThreadPoolExecutor
from typing import Union

from metrikflow.connectors.common.signals import add_signal_handler, handle_loop_stop
from metrikflow.logging import MetrikflowLogger
from metrikflow.metrics import MetricStore
from metrikflow.metrics.types import Event, Interval, Rate

from .s3_config import S3Config

try:
    import boto3
    has_connector = True

except Exception:
    boto3 = None
    has_connector = False


class S3:
    has_connector=has_connector

    def __init__(self, config: S3Config) -> None:
        self.aws_access_key_id = config.aws_access_key_id
        self.aws_secret_access_key = config.aws_secret_access_key
        self.region_name = config.region_name
        self.buckets_namespace = config.buckets_namespace

        self._executor = ThreadPoolExecutor(
            max_workers=os.cpu_count()
        )
        self.client = None
        self._loop = asyncio.get_event_loop()

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = MetrikflowLogger()
        self.logger.initialize()
        self._store = MetricStore()

    async def connect(self):

        add_signal_handler(
            self._loop,
            handle_loop_stop,
            self._executor,
            self._loop
        )

        self.client = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                boto3.client,
                's3',
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                region_name=self.region_name
            )
        )

    async def load(
        self,
        source_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60       
    ):
        
        try:
            
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_bucket,
                    Bucket=source_name,
                    CreateBucketConfiguration={
                        'LocationConstraint': self.region_name
                    }
                )
            )

        except Exception:
            pass
    
        metric_json = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.get_object,
                Bucket=source_name,
                Key=f'{metric.name}_{metric.metric_id}'
            )
        )

        return self._store.parse(
            json.loads(metric_json)
        )

    async def send(
        self,
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60       
    ):
        try:
            
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_bucket,
                    Bucket=destination_name,
                    CreateBucketConfiguration={
                        'LocationConstraint': self.region_name
                    }
                )
            )

        except Exception:
            pass

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.put_object,
                Bucket=destination_name,
                Key=f'{metric.name}_{metric.metric_id}',
                Body=metric.json()
            )
        )
        
    async def close(self):
        self._executor.shutdown(wait=False, cancel_futures=True)


import asyncio
from concurrent.futures import ThreadPoolExecutor


def handle_loop_stop(
    signame, 
    executor: ThreadPoolExecutor, 
    loop: asyncio.AbstractEventLoop,
): 
    try:
        executor.shutdown(wait=False, cancel_futures=True) 
        loop.stop()
    except Exception:
        pass




from datadog_api_client.v1.model.widget_layout import WidgetLayout
from typing import List, Union, Literal, Dict
from .size_type import (
    ExtraLargeSize,
    SmallSize,
    MediumSize,
    LargeSize,
    ExtraLargeSize
)
from .size_type_map import (
    SizeTypeMap,
    ExtraSmallSize,
    SmallSize,
    MediumSize,
    LargeSize,
    ExtraLargeSize
)


class Layout:

    def __init__(self) -> None:
        self.MAX_ROW_WIDTH=12
        self._current_row_width = 0
        self._layout_matrix: List[List[WidgetLayout]] = []
        self._widgets: Dict[str, WidgetLayout] = {}
        self._current_row: List[WidgetLayout] = []
        self._current_row_height = 0

    @property
    def widget_layout(self):
        return self._layout_matrix
    
    def __getitem__(self, widget_name: str):
        return self._widgets.get(widget_name)

    def add_widget(
        self, 
        widget_title: str,
        widget_size: Union[
            ExtraLargeSize,
            SmallSize,
            MediumSize,
            LargeSize,
            ExtraLargeSize
        ]
    ):
        requested_row_width = self._current_row_width + widget_size.WIDTH

        # If adding the widget would put us over the max width and we still haven't filled the row,
        # increase the size of the widgets in the row so that the row is filled.
        if requested_row_width > self.MAX_ROW_WIDTH and self._current_row_width < self.MAX_ROW_WIDTH:
            # Find the difference between the max row width and current row width
            # then distribute the difference among the existing widgets.
            difference = self.MAX_ROW_WIDTH - self._current_row_width

            row_items_count = len(self._current_row)
            last_item_width_buff = difference%row_items_count

            width_buffs = [
                int(difference/row_items_count) for _ in range(row_items_count)
            ]

            width_buffs[-1] += last_item_width_buff

            for item, width_buff in zip(
                self._current_row, 
                width_buffs
            ):
                item.width += width_buff

            self._append_new_row(
                widget_title,
                widget_size
            )

        # If adding the widget would put us over the max width and all widgets
        # completely fill the row.
        elif requested_row_width > self.MAX_ROW_WIDTH:
            self._append_new_row(
                widget_title,
                widget_size
            )

        else:
            widget_layout = WidgetLayout(
                height=widget_size.HEIGHT,
                width=widget_size.WIDTH,
                x=self._current_row_width,
                y=self._current_row_height
            )

            self._current_row.append(widget_layout)
            self._widgets[widget_title] = widget_layout

            self._current_row_width += widget_size.WIDTH
            

    def _append_new_row(
        self,
        widget_title: str,
        widget_size: Union[
            ExtraSmallSize,
            SmallSize,
            MediumSize,
            LargeSize,
            ExtraLargeSize
        ]
    ):
        row_height = self._get_row_max_height()
        self._current_row_height += row_height

        # Set all widget heights to the max height for uniformity's
        # sake
        for item in self._current_row:
            item.height = row_height

        # Append all widgets in the the current row to the matrix
        self._layout_matrix.extend(self._current_row)

        # Get the max height of the current row as the starting
        # height for the next row and initialize a new row
        # with the new widget

        widget_layout = WidgetLayout(
            height=widget_size.HEIGHT,
            width=widget_size.WIDTH,
            x=0,
            y=self._current_row_height
        )

        self._current_row = [widget_layout]

        # Set the starting x position of the next widget as the
        # width of the current widget.
        self._current_row_width = widget_size.WIDTH

        self._widgets[widget_title] = widget_layout

    def _get_row_max_height(self):
        # Get the max height of the current row
        return max([
            item.height for item in self._current_row
        ])


        

                




        
        



import uuid
import warnings
from typing import Any, Dict, List, Union

from metrikflow.logging import MetrikflowLogger
from metrikflow.metrics import MetricStore
from metrikflow.metrics.types import Event, Interval, Rate

from .mysql_config import MySQLConfig

try:
    # Aiomysql will raise warnings if a table exists despite us
    # explicitly passing "IF NOT EXISTS", so we're going to
    # ignore them.
    import aiomysql
    import sqlalchemy as sa
    warnings.filterwarnings('ignore', category=aiomysql.Warning)

    from aiomysql.sa import SAConnection, create_engine
    from sqlalchemy.schema import CreateTable
    
    has_connector = True

except Exception:
    sqlalchemy = object
    sa = object
    create_engine = object
    CreateTable = object
    SAConnection = object
    OperationalError = object
    has_connector = object



class MySQL:
    has_connector=has_connector

    def __init__(self, config: MySQLConfig) -> None:
        self.host = config.host
        self.database = config.database
        self.username = config.username
        self.password = config.password
        
        self._metrics_table = None

        self.metadata = sa.MetaData()
        self._engine = None
        self._connection = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = MetrikflowLogger()
        self.logger.initialize()

        self._store = MetricStore()

    async def connect(self):
        
        self._engine = await create_engine(
            db=self.database,
            host=self.host,
            user=self.username,
            password=self.password
        )

        self._connection: SAConnection = await self._engine.acquire()

    async def load(
        self,
        source_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],     
        timeout: Union[int, float]=None
    ):
        
        table = sa.Table(
            source_name,
            self.metadata,
            sa.Column('id', sa.Integer, primary_key=True),
            sa.Column('metric_id', sa.VARCHAR(255)),
            sa.Column('name', sa.VARCHAR(255)),
            sa.Column('kind', sa.VARCHAR(255)),
            sa.Column('group', sa.VARCHAR(255)),
            sa.Column('timestamp', sa.DateTime()),
            sa.Column('value', sa.Float()),
        )

        await self._connection.execute(
            CreateTable(
                table,
                if_not_exists=True
            )
        )

        metrics_data: List[Dict[str, Any]] = []

        async for row in self._connection.execute(
            table.select().where(table.c.name == metric.name).order_by(
                table.c.timestamp.desc()
            ).limit(1)
        ):
            metrics_data.append({
                'metric_id': row.metric_id,
                'name': row.name,
                'kind': row.kind,
                'group': row.group,
                'timestamp': row.timestamp,
                'value': row.value,
            })

        return self._store.parse(
            metrics_data[-1]
        )

    async def send(
        self,
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],     
        timeout: Union[int, float]=None
    ):
        table = sa.Table(
            destination_name,
            self.metadata,
            sa.Column('id', sa.Integer, primary_key=True),
            sa.Column('metric_id', sa.VARCHAR(255)),
            sa.Column('name', sa.VARCHAR(255)),
            sa.Column('kind', sa.VARCHAR(255)),
            sa.Column('group', sa.VARCHAR(255)),
            sa.Column('timestamp', sa.DateTime()),
            sa.Column('value', sa.Float()),
        )

        await self._connection.execute(
            CreateTable(
                table,
                if_not_exists=True
            )
        )

        async with self._connection.begin() as transaction:

            metric_data = metric.dict()
            metric_data['kind'] = metric.kind.value

            await self._connection.execute(
                table.insert().values(**metric_data)
            )

            await transaction.commit()

    async def close(self):
        await self._connection.close()
    





import uuid
import datetime
from pydantic import (
    BaseModel,
    StrictStr, 
    StrictInt,
    StrictFloat
)
from typing import Union, Dict, Any, List, Optional
from .metric_tag import MetricTag
from .metric_types import MetricTypes
from .unit_type import UnitType


class Metric(BaseModel):
    metric_id: StrictStr=str(uuid.uuid4())
    name: StrictStr
    kind: MetricTypes
    group: StrictStr
    timestamp: datetime.datetime
    value: Union[
            StrictInt, 
            StrictFloat
    ]
    tags: List[Dict[StrictStr, StrictStr]]=[]
    unit: Optional[UnitType]=None

    class Config:
        allow_arbitrary_values=True

    @classmethod
    def to_column_names(self):
        return [
            'metric_id',
            'name',
            'kind',
            'group',
            'timestamp',
            'value',
            'unit'
        ]

    @classmethod
    def parse(
        self,
        data: Dict[str, Any]
    ):
        raise Exception('Err. - Parse not implemented on base Metric class.')

    def update(
        self, 
        value: Union[StrictInt, 
            StrictFloat, 
            datetime.datetime, 
            StrictStr
        ]
    ):
        raise NotImplementedError('Err. - Update not implemented on base Metric class.')

    def to_date_string(self):
        return self.timestamp.strftime('%Y-%m-%dT%H:%M:%S.%f.%z')
    
    def serialize_value(self):
        return self.value

    def to_data_dict(self):
        
        data = {
            key: value for key, value in self.dict().items() if key != 'tags'
        }

        data['timestamp'] = self.to_date_string()
        data['kind'] = self.kind.value

        return data
    
    def to_tags_dict(self):
        tags = {
            key: value for key, value in self.dict().items() if key not in [
                'value', 
                'tags', 
                'timestamp',
                'metric_id'
            ]
        }

        tags['kind'] = self.kind.value

        for tag_set in self.tags:
            tags.update(tag_set)

        return tags
    
    def to_tags(self):
        tags: List[MetricTag] = []

        for tag in self.tags:
            tag_name, tag_value = list(tag.items())[0]

            tags.append(
                MetricTag(
                    name=tag_name,
                    value=tag_value
                )
            )

        return tags
    
    def get_tag(
        self,
        tag_name: str
    ):
        matching_tags = [
            tag for tag in self.tags if tag_name in tag
        ]

        if len(matching_tags) > 0:
            latest_matched_tag = matching_tags[-1]
            tag_name, tag_value = list(latest_matched_tag.items())[0]

            return MetricTag(
                name=tag_name,
                value=tag_value
            )
        
        else:
            return None
    


import uuid
from typing import Callable, Dict, Union

from metrikflow.logging import MetrikflowLogger
from metrikflow.metrics.types import Event, Interval, Rate
from metrikflow.metrics.types.base import MetricTypes

from .dogstatsd_config import DogStatsDConfig

try:
    from aio_statsd import DogStatsdClient

    from metrikflow.connectors.statsd import StatsD
    has_connector = True

except Exception:
    from metrikflow.connectors.empty import Empty as StatsD
    DogStatsdClient = None
    has_connector = False


class DogStatsD(StatsD):
    has_connector=has_connector

    def __init__(self, config: DogStatsDConfig) -> None:
        super(DogStatsD, self).__init__(config)
    
        self.host = config.host
        self.port = config.port

        self.connection = DogStatsdClient(
            host=self.host,
            port=self.port
        )

        self.stat_type_map = {
            MetricTypes.EVENT: 'gauge',
            MetricTypes.INTERVAL: 'gauge',
            MetricTypes.RATE: 'gauge'
        }

        self._update_map: Dict[
            str,
            Callable[
                [
                    str,
                    Union[int, float],
                    Union[int, float],
                    Dict[str, str]
                ],
                None
            ]
        ] = {
            'count': lambda: NotImplementedError('DogStatsD does not support counts.'),
            'gauge': self.connection.gauge,
            'sets': lambda: NotImplementedError('DogStatsD does not support sets.'),
            'increment': self.connection.increment,
            'histogram': self.connection.histogram,
            'distribution': self.connection.distribution,
            'timer': self.connection.timer
        }

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = MetrikflowLogger()
        self.logger.initialize()

        self.statsd_type = 'StatsD'

    async def load(
        self,
        source_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60
    ):
        raise Exception('Err. - DogStatsD connector is send only.')
        

    async def send(
        self,
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60
    ):
        
        metric_type = self.stat_type_map.get(
            metric.kind,
            'gauge'
        )

        metric_key = f'{destination_name}_{metric.name}'

        update_function = self._update_map.get(metric_type)
        update_function(
            metric_key,
            metric.value,
            sample_rate=1 if metric.kind == MetricTypes.RATE else None,
            tag_dict=metric.to_tags_dict()
        )

    async def close(self):
        await self.connection.close()


import asyncio
import datetime
import functools
import os
import uuid
from concurrent.futures import ThreadPoolExecutor
from typing import Union

from metrikflow.connectors.common.signals import add_signal_handler, handle_loop_stop
from metrikflow.logging import MetrikflowLogger
from metrikflow.metrics.types import Event, Interval, Rate

from .cloudwatch_config import CloudwatchConfig

try:
    import boto3
    has_connector = True

except Exception:
    boto3 = None
    has_connector = False
    

class Cloudwatch:
    has_connector=has_connector

    def __init__(self, config: CloudwatchConfig) -> None:
        self.aws_access_key_id = config.aws_access_key_id
        self.aws_secret_access_key = config.aws_secret_access_key
        self.region_name = config.region_name
        self.iam_role_arn = config.iam_role_arn
        self.schedule_rate = config.schedule_rate

        self.cloudwatch_targets = config.cloudwatch_targets
        self.aws_resource_arns = config.aws_resource_arns
        self.submit_timeout = config.submit_timeout

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = MetrikflowLogger()
        self.logger.initialize()

        self._executor = ThreadPoolExecutor(
            max_workers=os.cpu_count()
        )

        self.experiments_rule = None

        self.client = None
        self._loop = asyncio.get_event_loop()

    async def connect(self):

        add_signal_handler(
            self._loop,
            handle_loop_stop,
            self._executor,
            self._loop
        )

        self.client = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                boto3.client,
                'events',
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                region_name=self.region_name
            )
        )

    async def load(
        self,
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60
    ):
        raise Exception('Err. - CloudWatch connector is send only.')

    async def send(
        self,
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60
    ):
        await asyncio.wait_for(
            self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_events,
                    Entries=[
                        {
                            'Time': datetime.datetime.now(
                                datetime.timezone.utc
                            ),
                            'Detail': metric.json(),
                            'DetailType': destination_name,
                            'Resources': self.aws_resource_arns,
                            'Source': destination_name
                        }
                    ]
                )
            ),
            timeout=self.submit_timeout
        )

    async def close(self):
        self._executor.shutdown(wait=False, cancel_futures=True)


from .dashboard_type import DashboardType


class DashboardTypeMap:

    def __init__(self) -> None:
        self._types = {
            'datadog': DashboardType.DATADOG
        }

    def __iter__(self):
        for dashboard_type in self._types.values():
            yield dashboard_type

    def __getitem__(self, dashboard_type: str):
        return self._types.get(
            dashboard_type,
            DashboardType.DATADOG
        )
    
    def get(self, dashboard_type: str):
        return self._types.get(dashboard_type)


from typing import Dict
from metrik.connectors.common.types import ConnectorTypes
from pydantic import BaseModel


class DogStatsDConfig(BaseModel):
    host: str='localhost'
    port: int=8125
    custom_fields: Dict[str, str]={}
    reporter_type: ConnectorTypes=ConnectorTypes.DogStatsD


import asyncio
import functools
import os
import uuid
from concurrent.futures import ThreadPoolExecutor
from typing import Union

from metrikflow.connectors.common import ConnectorTypes
from metrikflow.connectors.common.signals import add_signal_handler, handle_loop_stop
from metrikflow.logging import MetrikflowLogger
from metrikflow.metrics import MetricStore
from metrikflow.metrics.types import Event, Interval, Rate

from .aws_lambda_config import AWSLambdaConfig

try:
    import boto3
    has_connector=True
except Exception:
    boto3 = None
    has_connector=False


class AWSLambda:
    has_connector=has_connector

    def __init__(self, config: AWSLambdaConfig) -> None:
        self.aws_access_key_id = config.aws_access_key_id
        self.aws_secret_access_key = config.aws_secret_access_key
        self.region_name = config.region_name

        self._executor = ThreadPoolExecutor(
            max_workers=os.cpu_count()
        )

        self._client = None
        self._loop = asyncio.get_event_loop()
        self.session_uuid = str(uuid.uuid4())

        self.reporter_type = ConnectorTypes.AWSLambda
        self.reporter_type_name = self.reporter_type.name.capitalize()
        self.metadata_string: str = None

        self.logger = MetrikflowLogger()
        self.logger.initialize()

        self._store = MetricStore()

    async def connect(self):

        add_signal_handler(
            self._loop,
            handle_loop_stop,
            self._executor,
            self._loop
        )

        self._client = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                boto3.client,
                'lambda',
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                region_name=self.region_name
            )
        )

    async def load(
        self,
        source_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60
    ):
        metric_json: Union[str, bytes] = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.invoke,
                FunctionName=source_name
            )
        )

        return self._store.parse(metric_json)

    async def send(
        self,
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60
    ):
        
        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.invoke,
                FunctionName=destination_name,
                Payload=metric.json()
            )
        )

    async def close(self):
        self._executor.shutdown(cancel_futures=True)
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')


import datetime
from pydantic import (
    BaseModel,
    StrictStr,
    StrictBool
)
from typing import Optional, Literal


class DatadogDashboardSummary(BaseModel):
    author_handle: StrictStr
    created_at: datetime.datetime
    deleted_at: Optional[datetime.datetime]=None
    description: Optional[StrictStr]=None
    id: StrictStr
    is_read_only: StrictBool
    layout_type: Literal[
        "ordered",
        "free"
    ]
    modified_at: datetime.datetime
    title: StrictStr
    url: StrictStr

    @classmethod
    def fields(cls):
        return [
            'author_handle',
            'created_at',
            'deleted_at',
            'description',
            'id',
            'is_read_only',
            'layout_type',
            'modified_at',
            'title',
            'url'
        ]


from typing import Dict, Optional
from pydantic import (
    BaseModel,
    StrictStr
)
from metrik.connectors.common.types import ConnectorTypes


class DatadogConfig(BaseModel):
    api_key: StrictStr
    app_key: StrictStr
    device_name: Optional[StrictStr]=None
    server_location: Optional[StrictStr]=None
    priority: StrictStr='normal'
    reporter_type: ConnectorTypes=ConnectorTypes.Datadog


from typing import Dict, Union, Literal
from .size_type import (
    SizeType,
    ExtraSmallSize,
    SmallSize,
    MediumSize,
    LargeSize,
    ExtraLargeSize
)


class SizeTypeMap:

    def __init__(self) -> None:
        self._types: Dict[
            str, 
            Union[
                ExtraSmallSize,
                SmallSize,
                MediumSize,
                LargeSize,
                ExtraLargeSize
            ]
        ] = {
            'extra_small': SizeType.EXTRA_SMALL.value,
            'small': SizeType.SMALL.value,
            'medium': SizeType.MEDIUM.value,
            'large': SizeType.LARGE.value,
            'extra_large': SizeType.EXTRA_LARGE.value
        }

        self._width_map: Dict[
            int, 
            Union[
                ExtraSmallSize,
                SmallSize,
                MediumSize,
                LargeSize,
                ExtraLargeSize
            ]
        ] = {
            3: SizeType.EXTRA_SMALL.value,
            4: SizeType.SMALL.value,
            6: SizeType.MEDIUM.value,
            9: SizeType.LARGE.value,
            12: SizeType.EXTRA_LARGE.value
        }

    def __getitem__(
        self, 
        size_type: Union[
            Literal['extra_small'],
            Literal['small'],
            Literal['medium'],
            Literal['large'],
            Literal['extra_large']
        ]
    ):
        return self._types.get(
            size_type,
            SizeType.MEDIUM.value
        )
    
    def get(
        self, 
        size_type: Union[
            Literal['extra_small'],
            Literal['small'],
            Literal['medium'],
            Literal['large'],
            Literal['extra_large']
        ]
    ):
        return self._types.get(size_type)
    
    def get_by_width(
        self, 
        width: Union[
            Literal[3],
            Literal[4],
            Literal[6],
            Literal[9],
            Literal[12]
        ]
    ):
        return self._width_map.get(width)
    

    


from collections import OrderedDict
from datadog_api_client.v1.model.formula_and_function_metric_query_definition import (
    FormulaAndFunctionMetricQueryDefinition,
)
from datadog_api_client.v1.model.widget import Widget
from metrik.dashboards.datadog.layouts import (
    Layout,
    SizeTypeMap
)
from metrik.dashboards.datadog.models import (
    WidgetMetric,
    Query,
    WidgetAggregationTypeMap,
    WidgetOptions,
    WidgetType,
    WidgetTypeMap
)
from .widget_types.change import ChangeFactory
from .widget_types.histogram import HistogramFactory
from .widget_types.query_value import QueryValueFactory
from .widget_types.scatter_plot import ScatterPlotFactory
from .widget_types.table import TableFactory
from .widget_types.timeseries import TimeseriesFactory
from typing import (
    List, 
    Dict,
    Optional,
    Union
)


class WidgetsFactory:

    def __init__(self) -> None:
        self._widgets: Dict[str, WidgetType] = OrderedDict()
        self._queries: Dict[str, FormulaAndFunctionMetricQueryDefinition]
        self._layouts = Layout()
        self._size_type_map = SizeTypeMap()
        self._widget_type_map = WidgetTypeMap()

        self._factories: Dict[
            WidgetType,
            Union[
                ChangeFactory,
                HistogramFactory,
                QueryValueFactory,
                ScatterPlotFactory,
                TableFactory,
                TimeseriesFactory,
            ]
        ] = {
            WidgetType.CHANGE: ChangeFactory(),
            WidgetType.HISTOGRAM: HistogramFactory(),
            WidgetType.QUERY_VALUE: QueryValueFactory(),
            WidgetType.SCATTER_PLOT: ScatterPlotFactory(),
            WidgetType.TABLE: TableFactory(),
            WidgetType.TIMESERIES: TimeseriesFactory()
        }

        self._aggregation_type_map = WidgetAggregationTypeMap()
        self._options: Dict[str, WidgetOptions] = {}

    def add_metric(
        self,
        widget_metric: WidgetMetric
    ):

        widget = self._widgets.get(widget_metric.widget_name)
        widget_type = self._widget_type_map[widget_metric.widget_type]
        
        if widget is None:

            self._widgets[widget_metric.widget_name] = widget_type
            widget_size = self._size_type_map[widget_metric.widget_size]

            self._options[widget_metric.widget_name] = widget_metric.widget_options

            self._layouts.add_widget(
                widget_metric.widget_name,
                widget_size
            )

            self._factories[widget_type].create_widget(
                widget_type,
                widget_size
            )

        metric_name = f'{widget_metric.metric.group}.{widget_metric.metric.name}'

        self._factories[widget_type].add_metric(
            widget_metric.widget_name,
            Query(
                aggregator=self._aggregation_type_map.map_to_datadog_type(
                    widget_metric.widget_aggregator
                ),
                metric=metric_name,
                aggregation_field=widget_metric.widget_aggregator_field,
                formula_type=self._aggregation_type_map.map_to_formula_type(
                    widget_metric.widget_aggregator
                )
            )
        )

    def to_widgets(
        self,
        options: Optional[WidgetOptions]=None
    ):
        
        widgets: List[Widget] = []
        default_options = WidgetOptions()

        for widget_name, widget_type in self._widgets.items():

            widget_options = self._options.get(widget_name)

            if widget_options is None and options is None:
                widget_options = default_options
      
            elif widget_options is None and options:
                widget_options = options

            elif widget_options and options:
                options_dict = options.dict(
                    exclude_defaults=True,
                    exclude_unset=True,
                    exclude_none=True
                )
                
                widget_options_dict = widget_options.dict(
                    exclude_defaults=True,
                    exclude_unset=True,
                    exclude_none=True
                )

                widget_options_dict.update(options_dict)

                widget_options = WidgetOptions(**widget_options_dict)
            
            widget = self._factories[widget_type].create(
                widget_name,
                widget_options
            )

            widgets.append(
                Widget(
                    definition=widget,
                    layout=self._layouts[widget_name]
                ),
            )

        return widgets
        



        
        

        




from .change import ChangeFactory
from .histogram import HistogramFactory
from .query_value import QueryValueFactory
from .scatter_plot import ScatterPlotFactory
from .table import TableFactory
from .timeseries import TimeseriesFactory


# # This is an ugly patch for: https://github.com/aio-libs/aiopg/issues/837
# import selectors  # isort:skip # noqa: F401

# selectors._PollLikeSelector.modify = (  # type: ignore
#     selectors._BaseSelectorImpl.modify  # type: ignore
# )

import uuid
from typing import Any, Dict, List, Union

from metrikflow.logging import MetrikflowLogger
from metrikflow.metrics import MetricStore
from metrikflow.metrics.types import Event, Interval, Rate

from .postgres_config import PostgresConfig

try:
    import sqlalchemy
    from sqlalchemy.dialects.postgresql import UUID
    from sqlalchemy.ext.asyncio import create_async_engine
    from sqlalchemy.ext.asyncio.engine import (
        AsyncConnection,
        AsyncEngine,
        AsyncTransaction,
    )
    from sqlalchemy.schema import CreateTable
    
    has_connector = True

except Exception:
    UUID = None
    sqlalchemy = None
    create_engine = None
    has_connector = False


class Postgres:
    has_connector=has_connector

    def __init__(self, config: PostgresConfig) -> None:
        self.host = config.host
        self.database = config.database
        self.username = config.username
        self.password = config.password

        self._engine = None
        self.metadata = sqlalchemy.MetaData()

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = MetrikflowLogger()
        self.logger.initialize()
        self.sql_type = 'Postgresql'

        self._store = MetricStore()

    async def connect(self):
  
        connection_uri = 'postgresql+asyncpg://'

        if self.username and self.password:
            connection_uri = f'{connection_uri}{self.username}:{self.password}@'

        self._engine: AsyncEngine = await create_async_engine(
            f'{connection_uri}{self.host}/{self.database}',
            echo=False
        )

    async def load(
        self,
        source_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60       
    ):
        
        table = sqlalchemy.Table(
            source_name,
            self.metadata,
            sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
            sqlalchemy.Column('metric_id', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('kind', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('group', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('timestamp', sqlalchemy.DateTime()),
            sqlalchemy.Column('value', sqlalchemy.Float()),
        )   


        metrics_data: List[Dict[str, Any]] = []

        async with self._engine.connect() as connection:
            connection: AsyncConnection = connection

            await connection.execute(
                CreateTable(
                    table,
                    if_not_exists=True
                )
            )

            async for row in connection.execute(
               table.select().where(
                  table.c.name == metric.name
               ).order_by(
                  table.c.timestamp.desc()
               ).limit(1)
            ):
                metrics_data.append({
                    'metric_id': row.metric_id,
                    'name': row.name,
                    'kind': row.kind,
                    'group': row.group,
                    'timestamp': row.timestamp,
                    'value': row.value
                })

        return self._store.parse(
            metrics_data[-1]
        )

    async def send(
        self,
        destination_name: str,
        metric: Union[
            Event,
            Interval,
            Rate
        ],
        timeout: Union[int, float]=60      
    ):

        table = sqlalchemy.Table(
            destination_name,
            self.metadata,
            sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
            sqlalchemy.Column('metric_id', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('kind', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('group', sqlalchemy.VARCHAR(255)),
            sqlalchemy.Column('timestamp', sqlalchemy.DateTime()),
            sqlalchemy.Column('value', sqlalchemy.Float())
        )   

        async with self._engine.connect() as connection:
            connection: AsyncConnection = connection

            await connection.execute(
                CreateTable(
                    table,
                    if_not_exists=True
                )
            )

            async with connection.begin() as transaction:
                transaction: AsyncTransaction = transaction

                metric_data = metric.dict()
                metric_data['kind'] = metric.kind.value

                await connection.execute(
                    table.insert().values(**metric_data)
                )

                await transaction.commit()

    async def close(self):
        pass



    



from metrik.metrics.types.base.unit_type import UnitType
from typing import Dict, Literal

class DatadogUnitMap:

    def __init__(self) -> None:
        self._unit_types: Dict[
            UnitType,
            Literal[
                'nanosecond',
                'microsecond',
                'millisecond',
                'second',
                'minute',
                'hour',
                'day',
                'week'
            ]
        ] = {
            UnitType.NANOSECONDS: 'nanosecond',
            UnitType.MICROSECONDS: 'microsecond',
            UnitType.MILLISECONDS: 'millisecond',
            UnitType.SECONDS: 'second',
            UnitType.MINUTES: 'minute',
            UnitType.HOURS: 'hour',
            UnitType.DAYS: 'day',
            UnitType.WEEKS: 'week'
        }

    def __getitem__(
        self,
        unit_type: UnitType
    ):
        return self._unit_types.get(
            unit_type,
            'second'
        )

    def get(
        self,
        unit_type: UnitType
    ):
        return self._unit_types.get(unit_type)


from datadog_api_client.v1.model.distribution_widget_definition import DistributionWidgetDefinition
from datadog_api_client.v1.model.distribution_widget_definition_type import DistributionWidgetDefinitionType
from datadog_api_client.v1.model.distribution_widget_histogram_request_type import (
    DistributionWidgetHistogramRequestType,
)
from datadog_api_client.v1.model.distribution_widget_request import DistributionWidgetRequest
from datadog_api_client.v1.model.distribution_widget_x_axis import DistributionWidgetXAxis
from datadog_api_client.v1.model.distribution_widget_y_axis import DistributionWidgetYAxis
from datadog_api_client.v1.model.formula_and_function_metric_query_definition import (
    FormulaAndFunctionMetricQueryDefinition,
)
from datadog_api_client.v1.model.formula_and_function_metric_data_source import (
    FormulaAndFunctionMetricDataSource,
)
from datadog_api_client.v1.model.widget_time import WidgetTime
from datadog_api_client.v1.model.widget_style import WidgetStyle
from datadog_api_client.v1.model.widget_text_align import WidgetTextAlign
from metrik.dashboards.datadog.models import (
    WidgetOptions,
    WidgetType
)
from metrik.dashboards.datadog.widgets.widget_types.base.base_widget_factory import BaseWidgetFactory


class HistogramFactory(BaseWidgetFactory[DistributionWidgetDefinition]):
    
    def __init__(self) -> None:
        super().__init__(WidgetType.HISTOGRAM)

    def create(
        self,
        widget_name: str,
        options: WidgetOptions
    ):
        
        queries = self.queries[widget_name]
        
        histogram = DistributionWidgetDefinition(
            title=widget_name,
            title_size="16",
            title_align=WidgetTextAlign.LEFT,
            show_legend=True,
            type=DistributionWidgetDefinitionType.DISTRIBUTION,
            legend_size='12',
            xaxis=DistributionWidgetXAxis(
                max="auto",
                include_zero=True,
                scale="linear",
                min="auto",
            ),
            yaxis=DistributionWidgetYAxis(
                max="auto",
                include_zero=True,
                scale="linear",
                min="auto",
            ),
            time=WidgetTime(
                live_span=options.to_datadog_live_span()
            ),
            requests=[
                DistributionWidgetRequest(
                    query=FormulaAndFunctionMetricQueryDefinition(
                        data_source=FormulaAndFunctionMetricDataSource.METRICS,
                        name=query.to_query_name(),
                        aggregator=query.formula_type,
                        query=query.to_query()
                    ) if query.aggregation_field else FormulaAndFunctionMetricQueryDefinition(
                        data_source=FormulaAndFunctionMetricDataSource.METRICS,
                        name=query.to_query_name(),
                        query=query.to_query()
                    ),
                    style=WidgetStyle(
                        palette="dog_classic",
                    ),
                )  for query in queries
            ],
        )

        self._widgets[widget_name] = histogram

        return histogram


from enum import Enum
from typing import TypedDict, Literal, Union


class MetricTypes(Enum):
    EVENT='EVENT'
    INTERVAL='INTERVAL'
    RATE='RATE'



class MetricTypesDict(TypedDict):
    event: Literal[MetricTypes.EVENT]
    interval: Literal[MetricTypes.INTERVAL]
    rate: Literal[MetricTypes.RATE]


class MetricTypesMap:

    def __init__(self) -> None:
        super().__init__()

        self._types: MetricTypesDict = {
            'event': MetricTypes.EVENT,
            'interval': MetricTypes.INTERVAL,
            'rate': MetricTypes.RATE
        }

    def __getitem__(
        self, 
        metric_type: Union[
            Literal['event'],
            Literal['interval'],
            Literal['rate']
        ]
    ):
        return self._types[metric_type]
    
    def get(
        self,
        metric_type: Union[
            Literal['event'],
            Literal['interval'],
            Literal['rate']
        ]
    ):
        if metric_type in self._types:
            return self._types[metric_type]
        
        return MetricTypes.EVENT



from typing import Dict
from pydantic import BaseModel
from metrik.connectors.common.types import ConnectorTypes


class StatsDConfig(BaseModel):
    host: str='localhost'
    port: int=8125
    reporter_type: ConnectorTypes=ConnectorTypes.StatsD